#!/usr/bin/env python3

"""
Functions for computing the grad B tensor and grad grad B tensor.
"""

import logging
import numpy as np
from .util import Struct, fourier_minimum

#logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

def calculate_grad_B_tensor(self):
    """
    Compute the components of the grad B tensor, and the scale
    length L grad B associated with the Frobenius norm of this
    tensor.
    The formula for the grad B tensor is eq (3.12) of
    Landreman (2021): Figures of merit for stellarators near the magnetic axis, JPP

    self should be an instance of qic with X1c, Y1s etc populated.
    """

    s = self # Shorthand
    tensor = Struct()
    
    factor = s.B0 * s.B0 / s.d_l_d_varphi / s.Bbar
    tensor.tn = s.sG * s.B0 * s.curvature
    tensor.nt = tensor.tn
    tensor.bb = factor * (s.X1c * s.d_Y1s_d_varphi - s.X1s * s.d_Y1c_d_varphi \
                          - s.iotaN * (s.X1s * s.Y1s + s.X1c * s.Y1c))
    tensor.nn = factor * (s.d_X1c_d_varphi * s.Y1s - s.d_X1s_d_varphi * s.Y1c \
                          + s.iotaN * (s.X1s * s.Y1s + s.X1c * s.Y1c))
    tensor.bn = factor * (-s.sG * s.Bbar * s.d_l_d_varphi * s.torsion / s.B0 \
                          + s.X1c * s.d_X1s_d_varphi - s.X1s * s.d_X1c_d_varphi \
                          - s.iotaN * (s.X1c * s.X1c + s.X1s * s.X1s))
    tensor.nb = factor * (s.d_Y1c_d_varphi * s.Y1s - s.d_Y1s_d_varphi * s.Y1c \
                          + s.sG * s.Bbar * s.d_l_d_varphi * s.torsion / s.B0 \
                          + s.iotaN * (s.Y1s * s.Y1s + s.Y1c * s.Y1c))
    if hasattr(s.B0, "__len__"): # check if B0 is an array (in quasisymmetry B0 is a scalar)
        tensor.tt = s.sG * np.matmul(s.d_d_varphi, s.B0) / s.d_l_d_varphi
    else:
        tensor.tt = 0

    self.grad_B_tensor = tensor
    
    t = s.tangent_cylindrical.transpose()
    n = s.normal_cylindrical.transpose()
    b = s.binormal_cylindrical.transpose()
    self.grad_B_tensor_cylindrical = np.array([[
                              tensor.nn * n[i] * n[j] \
                            + tensor.bn * b[i] * n[j] + tensor.nb * n[i] * b[j] \
                            + tensor.bb * b[i] * b[j] \
                            + tensor.tn * t[i] * n[j] + tensor.nt * n[i] * t[j] \
                            + tensor.tt * t[i] * t[j]
                        for i in range(3)] for j in range(3)])
    self.grad_B_tensor_cylindrical_array = np.reshape(self.grad_B_tensor_cylindrical, 9 * self.nphi)

    self.grad_B_colon_grad_B = tensor.tn * tensor.tn + tensor.nt * tensor.nt \
        + tensor.bb * tensor.bb + tensor.nn * tensor.nn \
        + tensor.nb * tensor.nb + tensor.bn * tensor.bn \
        + tensor.tt * tensor.tt

    self.L_grad_B = s.B0 * np.sqrt(2 / self.grad_B_colon_grad_B)
    self.inv_L_grad_B = 1.0 / self.L_grad_B
    self.min_L_grad_B = fourier_minimum(self.L_grad_B)
    
def calculate_grad_grad_B_tensor(self, two_ways=False):
    """
    Compute the components of the grad grad B tensor, and the scale
    length L grad grad B associated with the Frobenius norm of this
    tensor.
    self should be an instance of Qic with X1c, Y1s etc populated.
    The grad grad B tensor in discussed around eq (3.13)
    Landreman (2021): Figures of merit for stellarators near the magnetic axis, JPP
    although an explicit formula is not given there.

    If ``two_ways`` is ``True``, an independent calculation of
    the tensor is also computed, to confirm the answer is the same.
    """

    # Shortcuts
    s = self
    
    X1s = s.X1s
    X1c = s.X1c
    Y1s = s.Y1s
    Y1c = s.Y1c

    X20 = s.X20
    X2s = s.X2s
    X2c = s.X2c

    Y20 = s.Y20
    Y2s = s.Y2s
    Y2c = s.Y2c

    Z20 = s.Z20
    Z2s = s.Z2s
    Z2c = s.Z2c

    iota_N0 = s.iotaN
    iota = s.iota
    iota0 = iota
    lp = np.abs(s.G0) / s.B0

    curvature = s.curvature
    torsion = s.torsion
    d_l_d_varphi = self.d_l_d_varphi
    lprime = d_l_d_varphi

    sG = s.sG
    sign_G = s.sG
    sign_psi = s.spsi
    B0 = s.B0
    Bbar = s.Bbar
    G0 = s.G0
    I2 = s.I2
    G2 = s.G2
    p2 = s.p2

    B20 = s.B20
    B2s = s.B2s
    B2c = s.B2c

    d_B0_d_varphi = np.matmul(s.d_d_varphi, B0)

    d_X1s_d_varphi = s.d_X1s_d_varphi
    d_X1c_d_varphi = s.d_X1c_d_varphi
    d_Y1s_d_varphi = s.d_Y1s_d_varphi
    d_Y1c_d_varphi = s.d_Y1c_d_varphi

    d_X20_d_varphi = s.d_X20_d_varphi
    d_X2s_d_varphi = s.d_X2s_d_varphi
    d_X2c_d_varphi = s.d_X2c_d_varphi

    d_Y20_d_varphi = s.d_Y20_d_varphi
    d_Y2s_d_varphi = s.d_Y2s_d_varphi
    d_Y2c_d_varphi = s.d_Y2c_d_varphi

    d_Z20_d_varphi = s.d_Z20_d_varphi
    d_Z2s_d_varphi = s.d_Z2s_d_varphi
    d_Z2c_d_varphi = s.d_Z2c_d_varphi

    d2_B0_d_varphi2  = np.matmul(s.d_d_varphi, d_B0_d_varphi)
    d2_X1s_d_varphi2 = s.d2_X1s_d_varphi2
    d2_X1c_d_varphi2 = s.d2_X1c_d_varphi2
    d2_Y1s_d_varphi2 = s.d2_Y1s_d_varphi2
    d2_Y1c_d_varphi2 = s.d2_Y1c_d_varphi2
    d_curvature_d_varphi = s.d_curvature_d_varphi
    d_torsion_d_varphi = s.d_torsion_d_varphi

    grad_grad_B = np.zeros((s.nphi, 3, 3, 3))
    grad_grad_B_alt = np.zeros((s.nphi, 3, 3, 3))

    # The elements that follow are computed in the Mathematica notebook "20200407-01 Grad grad B tensor near axis"
    # and then formatted for fortran by the python script process_grad_grad_B_tensor_code

    # The order is (normal, binormal, tangent). So element 123 means nbt.

    # Element 111
    grad_grad_B[:,0,0,0] =(sG*B0**3*(8*iota_N0*X2c*Y1c* \
        Y1s + 4*iota_N0*X2s* \
        (-Y1c**2 + Y1s**2) -  \
        2*iota_N0*X1s*Y1c*Y20 +  \
        2*iota_N0*X1c*Y1s*Y20 +  \
        2*iota_N0*X1s*Y1c*Y2c +  \
        2*iota_N0*X1c*Y1s*Y2c -  \
        2*iota_N0*X1c*Y1c*Y2s +  \
        2*iota_N0*X1s*Y1s*Y2s -  \
        5*iota_N0*X1c*X1s*Y1c**2* \
        curvature +  \
        5*iota_N0*X1c**2*Y1c*Y1s* \
        curvature -  \
        5*iota_N0*X1s**2*Y1c*Y1s* \
        curvature +  \
        5*iota_N0*X1c*X1s*Y1s**2* \
        curvature -  \
        2*Y1c*Y20*d_X1c_d_varphi +  \
        2*Y1c*Y2c*d_X1c_d_varphi +  \
        2*Y1s*Y2s*d_X1c_d_varphi -  \
        5*X1s*Y1c*Y1s* \
        curvature*d_X1c_d_varphi +  \
        5*X1c*Y1s**2*curvature* \
        d_X1c_d_varphi -  \
        2*Y1s*Y20*d_X1s_d_varphi -  \
        2*Y1s*Y2c*d_X1s_d_varphi +  \
        2*Y1c*Y2s*d_X1s_d_varphi +  \
        5*X1s*Y1c**2*curvature* \
        d_X1s_d_varphi -  \
        5*X1c*Y1c*Y1s* \
        curvature*d_X1s_d_varphi +  \
        2*Y1c**2*d_X20_d_varphi +  \
        2*Y1s**2*d_X20_d_varphi -  \
        2*Y1c**2*d_X2c_d_varphi +  \
        2*Y1s**2*d_X2c_d_varphi -  \
        4*Y1c*Y1s*d_X2s_d_varphi))/ \
    (Bbar**2*lprime)

    # Element 112
    grad_grad_B[:,0,0,1] =(sG*B0**3*(-5*iota_N0*X1s* \
        Y1c**3*curvature +  \
        Y1c**2*(-6*iota_N0*Y2s +  \
            5*iota_N0*X1c*Y1s* \
            curvature +  \
            2*lprime*X20*torsion -  \
            2*lprime*X2c*torsion +  \
            5*lprime*X1s**2*curvature* \
            torsion +  \
            5*X1s*curvature* \
            d_Y1s_d_varphi +  \
            2*d_Y20_d_varphi -  \
            2*d_Y2c_d_varphi) +  \
        Y1s*(5*iota_N0*X1c*Y1s**2* \
            curvature -  \
            2*(lprime*(X1s* \
                    (Y20 + Y2c) -  \
                X1c*Y2s)*torsion -  \
                Y2s*d_Y1c_d_varphi +  \
                (Y20 + Y2c)* \
                d_Y1s_d_varphi) +  \
            Y1s*(6*iota_N0*Y2s +  \
                lprime*(2*X20 + 2*X2c +  \
                5*X1c**2*curvature)* \
                torsion +  \
                5*X1c*curvature* \
                d_Y1c_d_varphi +  \
                2*d_Y20_d_varphi +  \
                2*d_Y2c_d_varphi)) -  \
        Y1c*(5*iota_N0*X1s*Y1s**2* \
            curvature +  \
            2*lprime*(X1c* \
                (Y20 - Y2c) -  \
                X1s*Y2s)*torsion +  \
            2*Y20*d_Y1c_d_varphi -  \
            2*Y2c*d_Y1c_d_varphi -  \
            2*Y2s*d_Y1s_d_varphi +  \
            Y1s*(-12*iota_N0*Y2c +  \
                2*lprime*(2*X2s +  \
                5*X1c*X1s*curvature)* \
                torsion +  \
                5*X1s*curvature* \
                d_Y1c_d_varphi +  \
                5*X1c*curvature* \
                d_Y1s_d_varphi +  \
                4*d_Y2s_d_varphi))))/(Bbar**2*lprime)

    # Element 113
    grad_grad_B[:,0,0,2] =-((sG*B0**2*(Y1c**2* \
            (2*G2*sG + 2*I2*sG*iota0 - 4*B20*lprime +  \
            4*B2c*lprime +  \
            4*iota_N0*B0*Z2s +  \
            2*B0*lprime*X20* \
                curvature -  \
            2*B0*lprime*X2c* \
                curvature +  \
            B0*lprime*X1s**2* \
                curvature**2 +  \
            2*Z20*d_B0_d_varphi -  \
            2*Z2c*d_B0_d_varphi -  \
            2*B0*d_Z20_d_varphi +  \
            2*B0*d_Z2c_d_varphi) +  \
            Y1s*(2*B0*lprime* \
                (X1s*(Y20 + Y2c) -  \
                X1c*Y2s)*curvature +  \
            Y1s*(2*G2*sG + 2*I2*sG*iota0 -  \
                4*B20*lprime -  \
                4*B2c*lprime -  \
                4*iota_N0*B0*Z2s +  \
                2*B0*lprime*X20* \
                curvature +  \
                2*B0*lprime*X2c* \
                curvature +  \
                B0*lprime*X1c**2* \
                curvature**2 +  \
                2*Z20*d_B0_d_varphi +  \
                2*Z2c*d_B0_d_varphi -  \
                2*B0*d_Z20_d_varphi -  \
                2*B0*d_Z2c_d_varphi)) +  \
            2*Y1c*(4*B2s*lprime* \
                Y1s -  \
            2*Y1s*Z2s* \
                d_B0_d_varphi -  \
            B0*(lprime* \
                (X1c*(-Y20 + Y2c) +  \
                    X1s*Y2s)*curvature +  \
                Y1s* \
                (4*iota_N0*Z2c +  \
                    lprime*curvature* \
                    (2*X2s +  \
                        X1c*X1s*curvature) \
    - 2*d_Z2s_d_varphi)))))/(Bbar**2*lprime))

    # Element 121
    grad_grad_B[:,0,1,0] =(sG*B0**3*(3*iota_N0*X1s**3*Y1c* \
        curvature -  \
        3*iota_N0*X1c**3*Y1s* \
        curvature -  \
        3*X1s**2*curvature* \
        (iota_N0*X1c*Y1s +  \
            Y1c*(lprime*Y1c* \
                torsion - d_X1c_d_varphi)) +  \
        3*X1c**2*Y1s*curvature* \
        (-(lprime*Y1s*torsion) +  \
            d_X1s_d_varphi) -  \
        2*(lprime*(-2*X2s*Y1c* \
                Y1s +  \
                X2c* \
                (-Y1c**2 + Y1s**2) +  \
                X20*(Y1c**2 + Y1s**2))* \
            torsion +  \
            X2c*Y1c* \
            d_X1c_d_varphi +  \
            X2s*Y1s* \
            d_X1c_d_varphi +  \
            X2s*Y1c* \
            d_X1s_d_varphi -  \
            X2c*Y1s* \
            d_X1s_d_varphi -  \
            X20*(Y1c* \
                d_X1c_d_varphi +  \
                Y1s*d_X1s_d_varphi)) +  \
        X1s*(2*iota_N0*X20*Y1c -  \
            6*iota_N0*X2c*Y1c -  \
            6*iota_N0*X2s*Y1s +  \
            3*iota_N0*X1c**2*Y1c* \
            curvature +  \
            2*lprime*Y1s*Y20* \
            torsion +  \
            2*lprime*Y1s*Y2c* \
            torsion -  \
            2*lprime*Y1c*Y2s* \
            torsion +  \
            6*lprime*X1c*Y1c* \
            Y1s*curvature*torsion -  \
            3*X1c*Y1s*curvature* \
            d_X1c_d_varphi -  \
            3*X1c*Y1c*curvature* \
            d_X1s_d_varphi -  \
            2*Y1s*d_X20_d_varphi -  \
            2*Y1s*d_X2c_d_varphi +  \
            2*Y1c*d_X2s_d_varphi) +  \
        2*X1c*(3*iota_N0*X2s*Y1c -  \
            iota_N0*X20*Y1s -  \
            3*iota_N0*X2c*Y1s +  \
            lprime*Y1c*Y20* \
            torsion -  \
            lprime*Y1c*Y2c* \
            torsion -  \
            lprime*Y1s*Y2s* \
            torsion -  \
            Y1c*d_X20_d_varphi +  \
            Y1c*d_X2c_d_varphi +  \
            Y1s*d_X2s_d_varphi)))/\
    (Bbar**2*lprime)

    # Element 122
    grad_grad_B[:,0,1,1] =(sG*B0**3*(-4*iota_N0*X1s*Y1c* \
        Y2c - 4*iota_N0*X1c*Y1s* \
        Y2c + 4*iota_N0*X1c*Y1c* \
        Y2s - 4*iota_N0*X1s*Y1s* \
        Y2s + 3*iota_N0*X1c*X1s* \
        Y1c**2*curvature -  \
        3*iota_N0*X1c**2*Y1c*Y1s* \
        curvature +  \
        3*iota_N0*X1s**2*Y1c*Y1s* \
        curvature -  \
        3*iota_N0*X1c*X1s*Y1s**2* \
        curvature +  \
        2*X20*Y1c*d_Y1c_d_varphi +  \
        3*X1s**2*Y1c*curvature* \
        d_Y1c_d_varphi -  \
        3*X1c*X1s*Y1s* \
        curvature*d_Y1c_d_varphi -  \
        2*X2c*Y1c* \
        (2*iota_N0*Y1s + d_Y1c_d_varphi) +  \
        2*X20*Y1s*d_Y1s_d_varphi +  \
        2*X2c*Y1s*d_Y1s_d_varphi -  \
        3*X1c*X1s*Y1c* \
        curvature*d_Y1s_d_varphi +  \
        3*X1c**2*Y1s*curvature* \
        d_Y1s_d_varphi +  \
        2*X2s*(iota_N0*Y1c**2 -  \
            Y1s*(iota_N0*Y1s +  \
                d_Y1c_d_varphi) -  \
            Y1c*d_Y1s_d_varphi) -  \
        2*X1c*Y1c*d_Y20_d_varphi -  \
        2*X1s*Y1s*d_Y20_d_varphi +  \
        2*X1c*Y1c*d_Y2c_d_varphi -  \
        2*X1s*Y1s*d_Y2c_d_varphi +  \
        2*X1s*Y1c*d_Y2s_d_varphi +  \
        2*X1c*Y1s*d_Y2s_d_varphi))/ \
    (Bbar**2*lprime)

    # Element 123
    grad_grad_B[:,0,1,2] =(2*sG*B0**2*(X1s* \
        (2*B2s*lprime*Y1c +  \
            Y1s*(G2*sG + I2*sG*iota0 -  \
                2*B20*lprime -  \
                2*B2c*lprime -  \
                2*iota_N0*B0*Z2s +  \
                2*B0*lprime*X20* \
                curvature +  \
                2*B0*lprime*X2c* \
                curvature +  \
                Z20*d_B0_d_varphi +  \
                Z2c*d_B0_d_varphi -  \
                B0*d_Z20_d_varphi -  \
                B0*d_Z2c_d_varphi) -  \
            Y1c*(Z2s* \
                d_B0_d_varphi +  \
                B0*(2*iota_N0*Z2c +  \
                2*lprime*X2s*curvature -  \
                d_Z2s_d_varphi))) +  \
        X1c*(Y1c* \
            (G2*sG + I2*sG*iota0 - 2*B20*lprime +  \
                2*B2c*lprime +  \
                2*iota_N0*B0*Z2s +  \
                2*B0*lprime*X20* \
                curvature -  \
                2*B0*lprime*X2c* \
                curvature +  \
                Z20*d_B0_d_varphi -  \
                Z2c*d_B0_d_varphi -  \
                B0*d_Z20_d_varphi +  \
                B0*d_Z2c_d_varphi) +  \
            Y1s*(2*B2s*lprime -  \
                Z2s*d_B0_d_varphi +  \
                B0*(-2*iota_N0*Z2c -  \
                2*lprime*X2s*curvature +  \
                d_Z2s_d_varphi)))))/(Bbar**2*lprime)

    # Element 131
    grad_grad_B[:,0,2,0] =(sG*B0**2*(2*(-(X1s*Y1c) +  \
            X1c*Y1s)*d_B0_d_varphi* \
        (iota_N0*X1c*Y1c +  \
            iota_N0*X1s*Y1s +  \
            Y1s*d_X1c_d_varphi -  \
            Y1c*d_X1s_d_varphi) +  \
        B0*(lprime**2*curvature* \
            (-2*X2c*Y1c**2 -  \
                4*X2s*Y1c*Y1s +  \
                2*X2c*Y1s**2 +  \
                2*X20* \
                (Y1c**2 + Y1s**2) -  \
                2*X1c*Y1c*Y20 -  \
                2*X1s*Y1s*Y20 +  \
                2*X1c*Y1c*Y2c -  \
                2*X1s*Y1s*Y2c +  \
                2*X1s*Y1c*Y2s +  \
                2*X1c*Y1s*Y2s +  \
                3*X1s**2*Y1c**2* \
                curvature -  \
                6*X1c*X1s*Y1c* \
                Y1s*curvature +  \
                3*X1c**2*Y1s**2*curvature) \
    - Y1s**2*d_X1c_d_varphi**2 +  \
            iota_N0*X1c*Y1c**2* \
            d_X1s_d_varphi +  \
            iota_N0*X1c*Y1s**2* \
            d_X1s_d_varphi +  \
            2*Y1c*Y1s* \
            d_X1c_d_varphi*d_X1s_d_varphi -  \
            Y1c**2*d_X1s_d_varphi**2 +  \
            iota_N0*X1c**2*Y1s* \
            d_Y1c_d_varphi -  \
            X1c*Y1s*d_X1s_d_varphi* \
            d_Y1c_d_varphi -  \
            iota_N0*X1c**2*Y1c* \
            d_Y1s_d_varphi +  \
            X1c*Y1c*d_X1s_d_varphi* \
            d_Y1s_d_varphi +  \
            iota_N0*X1s**2* \
            (Y1s*d_Y1c_d_varphi -  \
                Y1c*d_Y1s_d_varphi) +  \
            lprime*(-(X1s*Y1c) +  \
                X1c*Y1s)*torsion* \
            (iota_N0*X1c**2 + iota_N0*X1s**2 -  \
                iota_N0*Y1c**2 - iota_N0*Y1s**2 +  \
                X1s*d_X1c_d_varphi -  \
                X1c*d_X1s_d_varphi -  \
                Y1s*d_Y1c_d_varphi +  \
                Y1c*d_Y1s_d_varphi) +  \
            X1c*Y1s**2* \
            d2_X1c_d_varphi2 -  \
            X1s*(Y1s* \
                d_X1c_d_varphi* \
                (iota_N0*Y1s - d_Y1c_d_varphi) \
    + Y1c*(d_X1c_d_varphi* \
                    d_Y1s_d_varphi +  \
                Y1s*d2_X1c_d_varphi2) +  \
                Y1c**2* \
                (iota_N0*d_X1c_d_varphi -  \
                d2_X1s_d_varphi2)) -  \
            X1c*Y1c*Y1s* \
            d2_X1s_d_varphi2)))/(Bbar**2*lprime**2)

    # Element 132
    grad_grad_B[:,0,2,1] =(sG*B0**2*((X1s*Y1c -  \
            X1c*Y1s)*d_B0_d_varphi* \
        (-2*iota_N0*Y1c**2 -  \
            Y1s*(2*iota_N0*Y1s +  \
                lprime*X1c*torsion +  \
                2*d_Y1c_d_varphi) +  \
            Y1c*(lprime*X1s* \
                torsion + 2*d_Y1s_d_varphi)) +  \
        B0*(iota_N0*Y1c**3* \
            d_X1s_d_varphi +  \
            lprime*(-(X1s*Y1c) +  \
                X1c*Y1s)* \
            (torsion* \
                (Y1s*d_X1c_d_varphi -  \
                Y1c*d_X1s_d_varphi) +  \
                X1s*(2*iota_N0*Y1s* \
                    torsion +  \
                torsion*d_Y1c_d_varphi -  \
                Y1c*d_torsion_d_varphi) +  \
                X1c*(2*iota_N0*Y1c* \
                    torsion -  \
                torsion*d_Y1s_d_varphi +  \
                Y1s*d_torsion_d_varphi)) +  \
            Y1s*(-(iota_N0*Y1s**2* \
                d_X1c_d_varphi) +  \
                d_Y1c_d_varphi* \
                (X1s*d_Y1c_d_varphi -  \
                X1c*d_Y1s_d_varphi) +  \
                Y1s*(iota_N0*X1s* \
                    d_Y1c_d_varphi -  \
                d_X1c_d_varphi* \
                    d_Y1c_d_varphi +  \
                X1c* \
                    (iota_N0*d_Y1s_d_varphi +  \
                    d2_Y1c_d_varphi2))) +  \
            Y1c*(iota_N0*Y1s**2* \
                d_X1s_d_varphi +  \
                d_Y1s_d_varphi* \
                (-(X1s*d_Y1c_d_varphi) +  \
                X1c*d_Y1s_d_varphi) +  \
                Y1s*(d_X1s_d_varphi* \
                    d_Y1c_d_varphi -  \
                2*iota_N0*X1s* \
                    d_Y1s_d_varphi +  \
                d_X1c_d_varphi* \
                    d_Y1s_d_varphi -  \
                X1s*d2_Y1c_d_varphi2 +  \
                X1c* \
                    (2*iota_N0*d_Y1c_d_varphi -  \
                    d2_Y1s_d_varphi2))) -  \
            Y1c**2*(iota_N0*Y1s* \
                d_X1c_d_varphi +  \
                (iota_N0*X1c + d_X1s_d_varphi)* \
                d_Y1s_d_varphi +  \
                X1s*(iota_N0*d_Y1c_d_varphi -  \
                d2_Y1s_d_varphi2)))))/ \
    (Bbar**2*lprime**2) 

    # Element 133 
    grad_grad_B[:,0,2,2] =(sG*B0**2*((-2*X2c*Y1c**2 -  \
            4*X2s*Y1c*Y1s +  \
            2*X2c*Y1s**2 +  \
            2*X20*(Y1c**2 + Y1s**2) -  \
            2*X1c*Y1c*Y20 -  \
            2*X1s*Y1s*Y20 +  \
            2*X1c*Y1c*Y2c -  \
            2*X1s*Y1s*Y2c +  \
            2*X1s*Y1c*Y2s +  \
            2*X1c*Y1s*Y2s +  \
            3*X1s**2*Y1c**2*curvature -  \
            6*X1c*X1s*Y1c* \
            Y1s*curvature +  \
            3*X1c**2*Y1s**2*curvature)* \
        d_B0_d_varphi +  \
        B0*(X1s*Y1c -  \
            X1c*Y1s)* \
        (iota_N0*X1s*Y1s*curvature +  \
            Y1s*curvature* \
            d_X1c_d_varphi -  \
            Y1c*curvature* \
            d_X1s_d_varphi +  \
            X1s*Y1c* \
            d_curvature_d_varphi +  \
            X1c*(iota_N0*Y1c* \
                curvature -  \
                Y1s*d_curvature_d_varphi))))/ \
    (Bbar**2*lprime) 

    # Element 211
    grad_grad_B[:,1,0,0] =(-2*sG*B0**3*(-(iota_N0*X1s**3* \
            Y1c*curvature) +  \
        X1s**2*(iota_N0*Y2s +  \
            curvature* \
            (iota_N0*X1c*Y1s +  \
                lprime*Y1c**2*torsion -  \
                Y1c*d_X1c_d_varphi)) +  \
        X1s*(2*iota_N0*X2c*Y1c +  \
            2*iota_N0*X2s*Y1s +  \
            2*iota_N0*X1c*Y2c -  \
            iota_N0*X1c**2*Y1c* \
            curvature -  \
            2*lprime*X1c*Y1c* \
            Y1s*curvature*torsion +  \
            Y2s*d_X1c_d_varphi +  \
            X1c*Y1s*curvature* \
            d_X1c_d_varphi -  \
            Y20*d_X1s_d_varphi -  \
            Y2c*d_X1s_d_varphi +  \
            X1c*Y1c*curvature* \
            d_X1s_d_varphi +  \
            Y1s*d_X20_d_varphi +  \
            Y1s*d_X2c_d_varphi -  \
            Y1c*d_X2s_d_varphi) +  \
        X1c*(-2*iota_N0*X2s*Y1c +  \
            2*iota_N0*X2c*Y1s -  \
            iota_N0*X1c*Y2s +  \
            iota_N0*X1c**2*Y1s* \
            curvature +  \
            lprime*X1c*Y1s**2* \
            curvature*torsion -  \
            Y20*d_X1c_d_varphi +  \
            Y2c*d_X1c_d_varphi +  \
            Y2s*d_X1s_d_varphi -  \
            X1c*Y1s*curvature* \
            d_X1s_d_varphi +  \
            Y1c*d_X20_d_varphi -  \
            Y1c*d_X2c_d_varphi -  \
            Y1s*d_X2s_d_varphi)))/ \
    (Bbar**2*lprime)

    # Element 212
    grad_grad_B[:,1,0,1] =(2*sG*B0**3*(X1s**2* \
        (lprime*(Y20 + Y2c)* \
            torsion +  \
            Y1c*curvature* \
            (iota_N0*Y1s + d_Y1c_d_varphi)) -  \
        X1s*(-(iota_N0*X1c*Y1c**2* \
                curvature) +  \
            iota_N0*X1c*Y1s**2* \
            curvature +  \
            2*lprime*X1c*Y2s* \
            torsion +  \
            Y2s*d_Y1c_d_varphi -  \
            Y20*d_Y1s_d_varphi -  \
            Y2c*d_Y1s_d_varphi +  \
            Y1s*(3*iota_N0*Y2s +  \
                lprime*(X20 + X2c)* \
                torsion +  \
                X1c*curvature* \
                d_Y1c_d_varphi +  \
                d_Y20_d_varphi +  \
                d_Y2c_d_varphi) +  \
            Y1c*(iota_N0*Y20 +  \
                3*iota_N0*Y2c -  \
                lprime*X2s*torsion +  \
                X1c*curvature* \
                d_Y1s_d_varphi -  \
                d_Y2s_d_varphi)) +  \
        X1c*(lprime*X1c*Y20* \
            torsion -  \
            lprime*X1c*Y2c* \
            torsion +  \
            Y20*d_Y1c_d_varphi -  \
            Y2c*d_Y1c_d_varphi -  \
            Y2s*d_Y1s_d_varphi +  \
            Y1c*(3*iota_N0*Y2s +  \
                lprime*(-X20 + X2c)* \
                torsion - d_Y20_d_varphi +  \
                d_Y2c_d_varphi) +  \
            Y1s*(iota_N0*Y20 -  \
                3*iota_N0*Y2c -  \
                iota_N0*X1c*Y1c* \
                curvature +  \
                lprime*X2s*torsion +  \
                X1c*curvature* \
                d_Y1s_d_varphi +  \
                d_Y2s_d_varphi))))/(Bbar**2*lprime)

    # Element 213
    grad_grad_B[:,1,0,2] =(2*sG*B0**2*(B0*lprime* \
        X1c**2*(Y20 - Y2c)* \
        curvature +  \
        X1s*(2*B2s*lprime* \
            Y1c - 2*iota_N0*B0*Y1c* \
            Z2c - B0*lprime* \
            X2s*Y1c*curvature +  \
            B0*lprime*X1s*Y20* \
            curvature +  \
            B0*lprime*X1s*Y2c* \
            curvature -  \
            Y1c*Z2s*d_B0_d_varphi +  \
            Y1s*(G2*sG + I2*sG*iota0 -  \
                2*B20*lprime -  \
                2*B2c*lprime -  \
                2*iota_N0*B0*Z2s +  \
                B0*lprime*X20* \
                curvature +  \
                B0*lprime*X2c* \
                curvature +  \
                Z20*d_B0_d_varphi +  \
                Z2c*d_B0_d_varphi -  \
                B0*d_Z20_d_varphi -  \
                B0*d_Z2c_d_varphi) +  \
            B0*Y1c*d_Z2s_d_varphi) +  \
        X1c*(2*B2s*lprime* \
            Y1s - 2*iota_N0*B0*Y1s* \
            Z2c - B0*lprime* \
            X2s*Y1s*curvature -  \
            2*B0*lprime*X1s* \
            Y2s*curvature -  \
            Y1s*Z2s*d_B0_d_varphi +  \
            Y1c*(G2*sG + I2*sG*iota0 -  \
                2*B20*lprime +  \
                2*B2c*lprime +  \
                2*iota_N0*B0*Z2s +  \
                B0*lprime*X20* \
                curvature -  \
                B0*lprime*X2c* \
                curvature +  \
                Z20*d_B0_d_varphi -  \
                Z2c*d_B0_d_varphi -  \
                B0*d_Z20_d_varphi +  \
                B0*d_Z2c_d_varphi) +  \
            B0*Y1s*d_Z2s_d_varphi)))/ \
    (Bbar**2*lprime)

    # Element 221
    grad_grad_B[:,1,1,0] =(-2*sG*B0**3*(X1c**2* \
        (3*iota_N0*X2s +  \
            lprime*(Y20 - Y2c)* \
            torsion - d_X20_d_varphi +  \
            d_X2c_d_varphi) +  \
        X1s*(lprime* \
            (X2s*Y1c -  \
                (X20 + X2c)*Y1s)* \
            torsion -  \
            X2s*d_X1c_d_varphi +  \
            X20*d_X1s_d_varphi +  \
            X2c*d_X1s_d_varphi -  \
            X1s*(3*iota_N0*X2s -  \
                lprime*(Y20 + Y2c)* \
                torsion + d_X20_d_varphi +  \
                d_X2c_d_varphi)) +  \
        X1c*(lprime* \
            (-(X20*Y1c) +  \
                X2c*Y1c +  \
                X2s*Y1s)*torsion +  \
            X20*d_X1c_d_varphi -  \
            X2c*d_X1c_d_varphi -  \
            X2s*d_X1s_d_varphi -  \
            2*X1s*(3*iota_N0*X2c +  \
                lprime*Y2s*torsion -  \
                d_X2s_d_varphi))))/(Bbar**2*lprime) \

    # Element 222 \
    grad_grad_B[:,1,1,1] =(2*sG*B0**3*(X1s**2* \
        (2*iota_N0*Y2s + d_Y20_d_varphi +  \
            d_Y2c_d_varphi) +  \
        X1c*(iota_N0*X2c*Y1s -  \
            2*iota_N0*X1c*Y2s +  \
            X2c*d_Y1c_d_varphi -  \
            X20*(iota_N0*Y1s +  \
                d_Y1c_d_varphi) +  \
            X2s*(-(iota_N0*Y1c) +  \
                d_Y1s_d_varphi) +  \
            X1c*d_Y20_d_varphi -  \
            X1c*d_Y2c_d_varphi) +  \
        X1s*(iota_N0*X2s*Y1s +  \
            4*iota_N0*X1c*Y2c +  \
            X2s*d_Y1c_d_varphi +  \
            X20*(iota_N0*Y1c -  \
                d_Y1s_d_varphi) +  \
            X2c*(iota_N0*Y1c -  \
                d_Y1s_d_varphi) -  \
            2*X1c*d_Y2s_d_varphi)))/ \
    (Bbar**2*lprime)

    # Element 223
    grad_grad_B[:,1,1,2] =(-2*sG*B0**2*(X1s**2* \
        (G2*sG + I2*sG*iota0 - 2*B20*lprime -  \
            2*B2c*lprime -  \
            2*iota_N0*B0*Z2s +  \
            2*B0*lprime*X20* \
            curvature +  \
            2*B0*lprime*X2c* \
            curvature +  \
            Z20*d_B0_d_varphi +  \
            Z2c*d_B0_d_varphi -  \
            B0*d_Z20_d_varphi -  \
            B0*d_Z2c_d_varphi) +  \
        X1c**2*(G2*sG + I2*sG*iota0 -  \
            2*B20*lprime +  \
            2*B2c*lprime +  \
            2*iota_N0*B0*Z2s +  \
            2*B0*lprime*X20* \
            curvature -  \
            2*B0*lprime*X2c* \
            curvature +  \
            Z20*d_B0_d_varphi -  \
            Z2c*d_B0_d_varphi -  \
            B0*d_Z20_d_varphi +  \
            B0*d_Z2c_d_varphi) +  \
        2*X1c*X1s* \
        (2*B2s*lprime -  \
            Z2s*d_B0_d_varphi +  \
            B0*(-2*iota_N0*Z2c -  \
                2*lprime*X2s*curvature +  \
                d_Z2s_d_varphi))))/(Bbar**2*lprime)

    # Element 231
    grad_grad_B[:,1,2,0] =(sG*B0**2*((X1s*Y1c -  \
            X1c*Y1s)*d_B0_d_varphi* \
        (2*iota_N0*X1c**2 +  \
            X1s*(2*iota_N0*X1s -  \
                lprime*Y1c*torsion +  \
                2*d_X1c_d_varphi) +  \
            X1c*(lprime*Y1s* \
                torsion - 2*d_X1s_d_varphi)) +  \
        B0*(2*lprime**2* \
            (X1c**2*(Y20 - Y2c) +  \
                X1s*(X2s*Y1c -  \
                X20*Y1s -  \
                X2c*Y1s +  \
                X1s*Y20 +  \
                X1s*Y2c) +  \
                X1c*(-(X20*Y1c) +  \
                X2c*Y1c +  \
                X2s*Y1s -  \
                2*X1s*Y2s))*curvature +  \
            iota_N0*X1c**3*d_Y1s_d_varphi +  \
            lprime*(-(X1s*Y1c) +  \
                X1c*Y1s)* \
            (torsion* \
                (Y1s*d_X1c_d_varphi -  \
                Y1c*d_X1s_d_varphi) +  \
                X1s*(2*iota_N0*Y1s* \
                    torsion +  \
                torsion*d_Y1c_d_varphi +  \
                Y1c*d_torsion_d_varphi) +  \
                X1c*(2*iota_N0*Y1c* \
                    torsion -  \
                torsion*d_Y1s_d_varphi -  \
                Y1s*d_torsion_d_varphi)) +  \
            X1s*(d_X1c_d_varphi* \
                (Y1s*d_X1c_d_varphi -  \
                Y1c*d_X1s_d_varphi) -  \
                iota_N0*X1s**2*d_Y1c_d_varphi +  \
                X1s*(iota_N0*Y1s* \
                    d_X1c_d_varphi -  \
                d_X1c_d_varphi* \
                    d_Y1c_d_varphi +  \
                Y1c* \
                    (iota_N0*d_X1s_d_varphi +  \
                    d2_X1c_d_varphi2))) +  \
            X1c*(d_X1s_d_varphi* \
                (-(Y1s*d_X1c_d_varphi) +  \
                Y1c*d_X1s_d_varphi) +  \
                iota_N0*X1s**2*d_Y1s_d_varphi +  \
                X1s*(d_X1s_d_varphi* \
                    d_Y1c_d_varphi +  \
                d_X1c_d_varphi* \
                    d_Y1s_d_varphi -  \
                Y1s* \
                    (2*iota_N0*d_X1s_d_varphi +  \
                    d2_X1c_d_varphi2) +  \
                Y1c* \
                    (2*iota_N0*d_X1c_d_varphi -  \
                    d2_X1s_d_varphi2))) -  \
            X1c**2*(iota_N0*Y1c* \
                d_X1s_d_varphi +  \
                iota_N0*X1s*d_Y1c_d_varphi +  \
                d_X1s_d_varphi* \
                d_Y1s_d_varphi +  \
                Y1s*(iota_N0*d_X1c_d_varphi -  \
                d2_X1s_d_varphi2)))))/ \
    (Bbar**2*lprime**2)

    # Element 232
    grad_grad_B[:,1,2,1] =-((sG*B0**2*(2* \
            (-(X1s*Y1c) +  \
            X1c*Y1s)* \
            d_B0_d_varphi* \
            (X1s*(iota_N0*Y1s +  \
                d_Y1c_d_varphi) +  \
            X1c*(iota_N0*Y1c -  \
                d_Y1s_d_varphi)) +  \
            B0*(lprime* \
                (-(X1s*Y1c) +  \
                X1c*Y1s)*torsion* \
                (iota_N0*X1c**2 +  \
                iota_N0*X1s**2 -  \
                iota_N0*Y1c**2 -  \
                iota_N0*Y1s**2 +  \
                X1s*d_X1c_d_varphi -  \
                X1c*d_X1s_d_varphi -  \
                Y1s*d_Y1c_d_varphi +  \
                Y1c*d_Y1s_d_varphi) +  \
            X1s**2* \
                (iota_N0*Y1s*d_Y1c_d_varphi +  \
                d_Y1c_d_varphi**2 -  \
                Y1c* \
                (iota_N0*d_Y1s_d_varphi +  \
                    d2_Y1c_d_varphi2)) +  \
            X1c*(iota_N0*Y1c**2* \
                d_X1s_d_varphi +  \
                iota_N0*Y1s**2* \
                d_X1s_d_varphi -  \
                Y1c* \
                (iota_N0*X1c +  \
                    d_X1s_d_varphi)* \
                d_Y1s_d_varphi +  \
                X1c*d_Y1s_d_varphi**2 +  \
                Y1s* \
                (iota_N0*X1c* \
                    d_Y1c_d_varphi +  \
                    d_X1s_d_varphi* \
                    d_Y1c_d_varphi -  \
                    X1c*d2_Y1s_d_varphi2)) +  \
            X1s*(-(iota_N0*Y1c**2* \
                    d_X1c_d_varphi) -  \
                iota_N0*Y1s**2* \
                d_X1c_d_varphi -  \
                2*X1c*d_Y1c_d_varphi* \
                d_Y1s_d_varphi +  \
                Y1s* \
                (-(d_X1c_d_varphi* \
                        d_Y1c_d_varphi) +  \
                    X1c*d2_Y1c_d_varphi2) +  \
                Y1c* \
                (d_X1c_d_varphi* \
                    d_Y1s_d_varphi +  \
                    X1c*d2_Y1s_d_varphi2)))))/ \
        (Bbar**2*lprime**2))

    # Element 233
    grad_grad_B[:,1,2,2] =(sG*B0**2*(2*(X1c**2* \
            (Y20 - Y2c) +  \
            X1s*(X2s*Y1c -  \
                X20*Y1s -  \
                X2c*Y1s +  \
                X1s*Y20 +  \
                X1s*Y2c) +  \
            X1c*(-(X20*Y1c) +  \
                X2c*Y1c +  \
                X2s*Y1s -  \
                2*X1s*Y2s))* \
        d_B0_d_varphi +  \
        B0*(-(X1s*Y1c) +  \
            X1c*Y1s)*curvature* \
        (iota_N0*X1c**2 +  \
            X1s*(iota_N0*X1s -  \
                2*lprime*Y1c*torsion +  \
                d_X1c_d_varphi) +  \
            X1c*(2*lprime*Y1s* \
                torsion - d_X1s_d_varphi))))/ \
    (Bbar**2*lprime) \

    # Element 311 \
    grad_grad_B[:,2,0,0] =-((sG*B0**2*(X1s*Y1c -  \
            X1c*Y1s)* \
        (3*sG**2*d_B0_d_varphi* \
            (iota_N0*X1c*Y1c +  \
            iota_N0*X1s*Y1s +  \
            Y1s*d_X1c_d_varphi -  \
            Y1c*d_X1s_d_varphi) +  \
            B0*(2*lprime**2* \
                (-(X1s*Y1c) +  \
                X1c*Y1s)*curvature**2 +  \
            sG**2*lprime*torsion* \
                (iota_N0*X1c**2 +  \
                iota_N0*X1s**2 -  \
                iota_N0*Y1c**2 -  \
                iota_N0*Y1s**2 +  \
                X1s*d_X1c_d_varphi -  \
                X1c*d_X1s_d_varphi -  \
                Y1s*d_Y1c_d_varphi +  \
                Y1c*d_Y1s_d_varphi) +  \
            sG**2*(iota_N0*X1c* \
                d_Y1c_d_varphi -  \
                d_X1s_d_varphi* \
                d_Y1c_d_varphi +  \
                iota_N0*X1s*d_Y1s_d_varphi +  \
                d_X1c_d_varphi* \
                d_Y1s_d_varphi +  \
                Y1s* \
                (iota_N0*d_X1s_d_varphi +  \
                    d2_X1c_d_varphi2) +  \
                Y1c* \
                (iota_N0*d_X1c_d_varphi -  \
                    d2_X1s_d_varphi2)))))/ \
        (Bbar**2*lprime**2))

    # Element 312
    grad_grad_B[:,2,0,1] =(sG*B0**2*(X1s*Y1c -  \
        X1c*Y1s)* \
        (-(d_B0_d_varphi* \
            (3*iota_N0*Y1c**2 +  \
            Y1s*(3*iota_N0*Y1s +  \
                2*lprime*X1c*torsion +  \
                3*d_Y1c_d_varphi) -  \
            Y1c*(2*lprime*X1s* \
                torsion + 3*d_Y1s_d_varphi))) -  \
        B0*(lprime* \
            (2*iota_N0*X1s*Y1s* \
                torsion +  \
                2*Y1s*torsion* \
                d_X1c_d_varphi -  \
                2*Y1c*torsion* \
                d_X1s_d_varphi -  \
                X1s*Y1c* \
                d_torsion_d_varphi +  \
                X1c*(2*iota_N0*Y1c* \
                    torsion +  \
                Y1s*d_torsion_d_varphi)) +  \
            Y1s*(2*iota_N0*d_Y1s_d_varphi +  \
                d2_Y1c_d_varphi2) +  \
            Y1c*(2*iota_N0*d_Y1c_d_varphi -  \
                d2_Y1s_d_varphi2))))/(Bbar**2*lprime**2) \

    # Element 313 \
    grad_grad_B[:,2,0,2] =(sG*B0**2*(X1s*Y1c -  \
        X1c*Y1s)* \
        ((2 + sG**2)*(X1s*Y1c -  \
            X1c*Y1s)*curvature* \
        d_B0_d_varphi +  \
        B0*((-1 + sG**2)*curvature* \
            (Y1s*d_X1c_d_varphi -  \
                Y1c*d_X1s_d_varphi) +  \
            X1s*(sG**2*iota_N0*Y1s* \
                curvature +  \
                curvature*d_Y1c_d_varphi +  \
                Y1c*d_curvature_d_varphi) +  \
            X1c*(sG**2*iota_N0*Y1c* \
                curvature -  \
                curvature*d_Y1s_d_varphi -  \
                Y1s*d_curvature_d_varphi))))/ \
    (Bbar**2*lprime)

    # Element 321
    grad_grad_B[:,2,1,0] =-((sG*B0**2*(X1s*Y1c -  \
            X1c*Y1s)* \
        (-(d_B0_d_varphi* \
            (3*iota_N0*X1c**2 +  \
                X1s*(3*iota_N0*X1s -  \
                    2*lprime*Y1c*torsion +  \
                    3*d_X1c_d_varphi) +  \
                X1c*(2*lprime*Y1s* \
                    torsion - 3*d_X1s_d_varphi))) \
    + B0*(lprime* \
                (X1s*(2*iota_N0*Y1s* \
                    torsion +  \
                    2*torsion*d_Y1c_d_varphi +  \
                    Y1c*d_torsion_d_varphi) +  \
                X1c* \
                (2*iota_N0*Y1c*torsion -  \
                    2*torsion*d_Y1s_d_varphi -  \
                    Y1s*d_torsion_d_varphi)) -  \
            X1s*(2*iota_N0*d_X1s_d_varphi +  \
                d2_X1c_d_varphi2) +  \
            X1c*(-2*iota_N0*d_X1c_d_varphi +  \
                d2_X1s_d_varphi2))))/(Bbar**2*lprime**2) \
    )

    # Element 322
    grad_grad_B[:,2,1,1] =(sG*B0**2*(X1s*Y1c -  \
        X1c*Y1s)* \
        (3*d_B0_d_varphi* \
        (X1s*(iota_N0*Y1s +  \
                d_Y1c_d_varphi) +  \
            X1c*(iota_N0*Y1c -  \
                d_Y1s_d_varphi)) +  \
        B0*(iota_N0*Y1c* \
            d_X1c_d_varphi +  \
            iota_N0*Y1s*d_X1s_d_varphi +  \
            iota_N0*X1c*d_Y1c_d_varphi +  \
            d_X1s_d_varphi*d_Y1c_d_varphi +  \
            iota_N0*X1s*d_Y1s_d_varphi -  \
            d_X1c_d_varphi*d_Y1s_d_varphi +  \
            lprime*torsion* \
            (iota_N0*X1c**2 + iota_N0*X1s**2 -  \
                iota_N0*Y1c**2 - iota_N0*Y1s**2 +  \
                X1s*d_X1c_d_varphi -  \
                X1c*d_X1s_d_varphi -  \
                Y1s*d_Y1c_d_varphi +  \
                Y1c*d_Y1s_d_varphi) +  \
            X1s*d2_Y1c_d_varphi2 -  \
            X1c*d2_Y1s_d_varphi2)))/ \
    (Bbar**2*lprime**2)

    # Element 323
    grad_grad_B[:,2,1,2] =-((sG*B0**3*(X1s*Y1c -  \
            X1c*Y1s)*curvature* \
        (sG**2*iota_N0*X1c**2 +  \
            X1s*(sG**2*iota_N0*X1s -  \
            (1 + sG**2)*lprime*Y1c* \
                torsion + sG**2*d_X1c_d_varphi) +  \
            X1c*((1 + sG**2)*lprime*Y1s* \
                torsion - sG**2*d_X1s_d_varphi)))/ \
        (Bbar**2*lprime)) \

    # Element 331
    grad_grad_B[:,2,2,0] =(sG*B0**2*(X1s*Y1c -  \
        X1c*Y1s)* \
        ((2 + sG**2)*(X1s*Y1c -  \
            X1c*Y1s)*curvature* \
        d_B0_d_varphi +  \
        B0*((-1 + sG**2)*curvature* \
            (Y1s*d_X1c_d_varphi -  \
                Y1c*d_X1s_d_varphi) +  \
            X1s*(sG**2*iota_N0*Y1s* \
                curvature +  \
                curvature*d_Y1c_d_varphi +  \
                Y1c*d_curvature_d_varphi) +  \
            X1c*(sG**2*iota_N0*Y1c* \
                curvature -  \
                curvature*d_Y1s_d_varphi -  \
                Y1s*d_curvature_d_varphi))))/ \
    (Bbar**2*lprime)

    # Element 332
    grad_grad_B[:,1,1,0] =(sG*B0**3*(X1s*Y1c -  \
        X1c*Y1s)*curvature* \
        (sG**2*iota_N0*Y1c**2 +  \
        Y1s*(sG**2*iota_N0*Y1s +  \
            (-1 + sG**2)*lprime*X1c* \
            torsion + sG**2*d_Y1c_d_varphi) -  \
        Y1c*((-1 + sG**2)*lprime*X1s* \
            torsion + sG**2*d_Y1s_d_varphi)))/ \
    (Bbar**2*lprime)

    # Element 333
    grad_grad_B[:,1,1,1] =(sG*B0*(X1s*Y1c -  \
        X1c*Y1s)* \
        (2*B0**2*lprime**2* \
        (-(X1s*Y1c) +  \
            X1c*Y1s)*curvature**2 +  \
        2*sG**2*(X1s*Y1c -  \
            X1c*Y1s)* \
        d_B0_d_varphi**2 +  \
        sG**2*B0*(d_B0_d_varphi* \
            (X1s*d_Y1c_d_varphi -  \
                X1c*d_Y1s_d_varphi) -  \
            Y1s*(d_B0_d_varphi* \
                d_X1c_d_varphi +  \
                X1c*d2_B0_d_varphi2) +  \
            Y1c*(d_B0_d_varphi* \
                d_X1s_d_varphi +  \
                X1s*d2_B0_d_varphi2))))/ \
    (Bbar**2*lprime**2) \



    self.grad_grad_B = grad_grad_B

    # Compute the (inverse) scale length
    squared = grad_grad_B * grad_grad_B
    norm_squared = np.sum(squared, axis=(1,2,3))
    self.grad_grad_B_inverse_scale_length_vs_varphi = np.sqrt(np.sqrt(norm_squared) / (4*B0))
    self.L_grad_grad_B = 1 / self.grad_grad_B_inverse_scale_length_vs_varphi
    self.grad_grad_B_inverse_scale_length = np.max(self.grad_grad_B_inverse_scale_length_vs_varphi)

    if not two_ways:
        return

    # Build the whole tensor again using Rogerio's approach,
    # "20200424-01 Rogerio's GradGradB calculation.nb"
    # and verify the two calculations match.

    # Element 111
    grad_grad_B_alt[:,0,0,0] =(-2*B0*(-4*sign_G*sign_psi*iota_N0*X2c*Y1c*\
                                      Y1s + iota_N0*X1c*X1c*\
                                      Y1c*(Y1c*\
                                           (-Y20 + Y2c) + \
                                           Y1s*(Y2s - \
                                                2*sign_G*sign_psi*curvature)) + \
                                      X20*Y1c*Y1c*Y1s*\
                                      d_X1c_d_varphi - \
                                      X2c*Y1c*Y1c*Y1s*\
                                      d_X1c_d_varphi + \
                                      X20*Y1s*Y1s*Y1s*\
                                      d_X1c_d_varphi + \
                                      X2c*Y1s*Y1s*Y1s*\
                                      d_X1c_d_varphi + \
                                      sign_G*sign_psi*Y1c*Y20*\
                                      d_X1c_d_varphi - \
                                      sign_G*sign_psi*Y1c*Y2c*\
                                      d_X1c_d_varphi - \
                                      sign_G*sign_psi*Y1s*Y2s*\
                                      d_X1c_d_varphi - \
                                      2*X2s*(sign_G*sign_psi*iota_N0*Y1s*Y1s + \
                                             Y1c*Y1c*\
                                             (-(sign_G*sign_psi*iota_N0) + \
                                              iota_N0*X1c*Y1s) + \
                                             Y1c*Y1s*Y1s*\
                                             d_X1c_d_varphi) + \
                                      X1c*(iota_N0*X2c*Y1c*\
                                           (-Y1c*Y1c + Y1s*Y1s) + \
                                           iota_N0*X20*Y1c*\
                                           (Y1c*Y1c + Y1s*Y1s) - \
                                           sign_G*sign_psi*iota_N0*Y1s*Y20 - \
                                           sign_G*sign_psi*iota_N0*Y1s*Y2c + \
                                           sign_G*sign_psi*iota_N0*Y1c*Y2s - \
                                           Y1c*Y1s*Y20*\
                                           d_X1c_d_varphi + \
                                           Y1c*Y1s*Y2c*\
                                           d_X1c_d_varphi + \
                                           Y1s*Y1s*Y2s*\
                                           d_X1c_d_varphi - \
                                           2*sign_G*sign_psi*Y1s*Y1s*curvature*\
                                           d_X1c_d_varphi) - \
                                      sign_G*sign_psi*Y1c*Y1c*d_X20_d_varphi - \
                                      sign_G*sign_psi*Y1s*Y1s*d_X20_d_varphi + \
                                      sign_G*sign_psi*Y1c*Y1c*d_X2c_d_varphi - \
                                      sign_G*sign_psi*Y1s*Y1s*d_X2c_d_varphi + \
                                      2*sign_G*sign_psi*Y1c*Y1s*\
                                      d_X2s_d_varphi))/(lp*sign_psi)

    # Element 112
    grad_grad_B_alt[:,0,0,1] =(2*B0*(2*iota_N0*X2s*Y1c*Y1c*Y1c*\
                                     Y1s + 2*iota_N0*X2s*\
                                     Y1c*Y1s*Y1s*Y1s + \
                                     iota_N0*X1c*Y1c*Y1c*Y1c*\
                                     Y20 + iota_N0*X1c*Y1c*\
                                     Y1s*Y1s*Y20 - \
                                     iota_N0*X1c*Y1c*Y1c*Y1c*\
                                     Y2c + 6*sign_G*sign_psi*iota_N0*Y1c*\
                                     Y1s*Y2c - \
                                     iota_N0*X1c*Y1c*Y1s*Y1s*\
                                     Y2c - 3*sign_G*sign_psi*iota_N0*Y1c*Y1c*\
                                     Y2s - iota_N0*X1c*\
                                     Y1c*Y1c*Y1s*Y2s + \
                                     3*sign_G*sign_psi*iota_N0*Y1s*Y1s*Y2s - \
                                     iota_N0*X1c*Y1s*Y1s*Y1s*\
                                     Y2s + 2*sign_G*sign_psi*iota_N0*X1c*\
                                     Y1c*Y1c*Y1s*curvature + \
                                     2*sign_G*sign_psi*iota_N0*X1c*Y1s*Y1s*Y1s*\
                                     curvature - \
                                     2*lp*sign_G*sign_psi*X2s*Y1c*\
                                     Y1s*torsion + \
                                     2*lp*X1c*X2s*Y1c*\
                                     Y1s*Y1s*torsion - \
                                     lp*sign_G*sign_psi*X1c*Y1c*\
                                     Y20*torsion + \
                                     lp*X1c*X1c*Y1c*Y1s*\
                                     Y20*torsion + \
                                     lp*sign_G*sign_psi*X1c*Y1c*\
                                     Y2c*torsion - \
                                     lp*X1c*X1c*Y1c*Y1s*\
                                     Y2c*torsion + \
                                     lp*sign_G*sign_psi*X1c*Y1s*\
                                     Y2s*torsion - \
                                     lp*X1c*X1c*Y1s*Y1s*Y2s*\
                                     torsion + \
                                     2*lp*sign_G*sign_psi*X1c*X1c*Y1s*Y1s*\
                                     curvature*torsion + \
                                     2*X2s*Y1c*Y1s*Y1s*\
                                     d_Y1c_d_varphi - \
                                     sign_G*sign_psi*Y1c*Y20*\
                                     d_Y1c_d_varphi + \
                                     X1c*Y1c*Y1s*\
                                     Y20*d_Y1c_d_varphi + \
                                     sign_G*sign_psi*Y1c*Y2c*\
                                     d_Y1c_d_varphi - \
                                     X1c*Y1c*Y1s*\
                                     Y2c*d_Y1c_d_varphi + \
                                     sign_G*sign_psi*Y1s*Y2s*\
                                     d_Y1c_d_varphi - \
                                     X1c*Y1s*Y1s*Y2s*\
                                     d_Y1c_d_varphi + \
                                     2*sign_G*sign_psi*X1c*Y1s*Y1s*\
                                     curvature*d_Y1c_d_varphi - \
                                     2*X2s*Y1c*Y1c*Y1s*\
                                     d_Y1s_d_varphi - \
                                     X1c*Y1c*Y1c*Y20*\
                                     d_Y1s_d_varphi - \
                                     sign_G*sign_psi*Y1s*Y20*\
                                     d_Y1s_d_varphi + \
                                     X1c*Y1c*Y1c*Y2c*\
                                     d_Y1s_d_varphi - \
                                     sign_G*sign_psi*Y1s*Y2c*\
                                     d_Y1s_d_varphi + \
                                     sign_G*sign_psi*Y1c*Y2s*\
                                     d_Y1s_d_varphi + \
                                     X1c*Y1c*Y1s*\
                                     Y2s*d_Y1s_d_varphi - \
                                     2*sign_G*sign_psi*X1c*Y1c*Y1s*\
                                     curvature*d_Y1s_d_varphi + \
                                     X2c*(Y1c*Y1c - Y1s*Y1s)*\
                                     (iota_N0*Y1c*Y1c + iota_N0*Y1s*Y1s - \
                                      lp*sign_G*sign_psi*torsion + \
                                      Y1s*(lp*X1c*torsion + \
                                           d_Y1c_d_varphi) - \
                                      Y1c*d_Y1s_d_varphi) - \
                                     X20*(Y1c*Y1c + Y1s*Y1s)*\
                                     (iota_N0*Y1c*Y1c + iota_N0*Y1s*Y1s - \
                                      lp*sign_G*sign_psi*torsion + \
                                      Y1s*(lp*X1c*torsion + \
                                           d_Y1c_d_varphi) - \
                                      Y1c*d_Y1s_d_varphi) + \
                                     sign_G*sign_psi*Y1c*Y1c*d_Y20_d_varphi + \
                                     sign_G*sign_psi*Y1s*Y1s*d_Y20_d_varphi - \
                                     sign_G*sign_psi*Y1c*Y1c*d_Y2c_d_varphi + \
                                     sign_G*sign_psi*Y1s*Y1s*d_Y2c_d_varphi - \
                                     2*sign_G*sign_psi*Y1c*Y1s*\
                                     d_Y2s_d_varphi))/(lp*sign_psi)

    # Element 113
    grad_grad_B_alt[:,0,0,2] =(-2*(Y1c*Y1c*(G2*sign_psi + I2*sign_psi*iota - \
                                            2*lp*sign_G*sign_psi*B20 + \
                                            2*lp*sign_G*sign_psi*B2c + \
                                            2*B0*sign_G*sign_psi*iota_N0*Z2s + \
                                            B0*lp*sign_G*sign_psi*X20*curvature - \
                                            B0*lp*sign_G*sign_psi*X2c*curvature + \
                                            B0*lp*X1c*X20*Y1s*\
                                            curvature - \
                                            B0*lp*X1c*X2c*Y1s*\
                                            curvature - \
                                            B0*sign_G*sign_psi*d_Z20_d_varphi + \
                                            B0*sign_G*sign_psi*d_Z2c_d_varphi) + \
                                   Y1s*(B0*lp*X1c*\
                                        (X20 + X2c)*Y1s*Y1s*\
                                        curvature - \
                                        B0*lp*sign_G*sign_psi*X1c*Y2s*\
                                        curvature + \
                                        Y1s*(G2*sign_psi + I2*sign_psi*iota - \
                                             2*lp*sign_G*sign_psi*B20 - \
                                             2*lp*sign_G*sign_psi*B2c - \
                                             2*B0*sign_G*sign_psi*iota_N0*Z2s + \
                                             B0*lp*sign_G*sign_psi*X20*curvature + \
                                             B0*lp*sign_G*sign_psi*X2c*curvature + \
                                             B0*lp*X1c*X1c*Y2s*\
                                             curvature + \
                                             B0*lp*sign_G*sign_psi*X1c*X1c*\
                                             curvature*curvature - \
                                             B0*sign_G*sign_psi*d_Z20_d_varphi - \
                                             B0*sign_G*sign_psi*d_Z2c_d_varphi)) + \
                                   Y1c*(4*lp*sign_G*sign_psi*B2s*\
                                        Y1s - \
                                        B0*(2*lp*X1c*X2s*\
                                            Y1s*Y1s*curvature + \
                                            lp*sign_G*sign_psi*X1c*\
                                            (-Y20 + Y2c)*\
                                            curvature + \
                                            Y1s*\
                                            (4*sign_G*sign_psi*iota_N0*Z2c + \
                                             2*lp*sign_G*sign_psi*X2s*curvature + \
                                             lp*X1c*X1c*Y20*\
                                             curvature - \
                                             lp*X1c*X1c*Y2c*\
                                             curvature - \
                                             2*sign_G*sign_psi*d_Z2s_d_varphi)))))/\
                                             (lp*sign_psi)

    # Element 121
    grad_grad_B_alt[:,0,1,0] =(-2*B0*(iota_N0*X1c*X1c*X1c*\
                                      (Y1c*(Y20 - Y2c) + \
                                       Y1s*(-Y2s + \
                                            sign_G*sign_psi*curvature)) - \
                                      X1c*X1c*(iota_N0*X2c*\
                                               (-Y1c*Y1c + Y1s*Y1s) + \
                                               iota_N0*X20*\
                                               (Y1c*Y1c + Y1s*Y1s) + \
                                               Y1s*(-2*iota_N0*X2s*\
                                                    Y1c + \
                                                    lp*(Y1c*\
                                                        (-Y20 + Y2c) + \
                                                        Y1s*\
                                                        (Y2s - sign_G*sign_psi*curvature))*\
                                                    torsion)) + \
                                      sign_G*sign_psi*(X2s*Y1s*\
                                                       (-2*lp*Y1c*torsion + \
                                                        d_X1c_d_varphi) + \
                                                       X20*(lp*Y1c*Y1c*\
                                                            torsion + \
                                                            lp*Y1s*Y1s*torsion - \
                                                            Y1c*d_X1c_d_varphi) + \
                                                       X2c*(-(lp*Y1c*Y1c*\
                                                              torsion) + \
                                                            lp*Y1s*Y1s*torsion + \
                                                            Y1c*d_X1c_d_varphi)) + \
                                      X1c*(3*sign_G*sign_psi*iota_N0*X2c*\
                                           Y1s + \
                                           lp*X2c*Y1c*Y1c*Y1s*\
                                           torsion - \
                                           lp*X2c*Y1s*Y1s*Y1s*\
                                           torsion - \
                                           lp*sign_G*sign_psi*Y1c*Y20*\
                                           torsion + \
                                           lp*sign_G*sign_psi*Y1c*Y2c*\
                                           torsion + \
                                           lp*sign_G*sign_psi*Y1s*Y2s*\
                                           torsion - \
                                           X20*Y1s*\
                                           (-(sign_G*sign_psi*iota_N0) + \
                                            lp*Y1c*Y1c*torsion + \
                                            lp*Y1s*Y1s*torsion) + \
                                           X2s*Y1c*\
                                           (-3*sign_G*sign_psi*iota_N0 + \
                                            2*lp*Y1s*Y1s*torsion) + \
                                           sign_G*sign_psi*Y1c*d_X20_d_varphi - \
                                           sign_G*sign_psi*Y1c*d_X2c_d_varphi - \
                                           sign_G*sign_psi*Y1s*d_X2s_d_varphi)))/\
                                           (lp*sign_psi)

    # Element 122
    grad_grad_B_alt[:,0,1,1] =(2*B0*(-(X1c*X1c*\
                                       (Y1c*(Y20 - Y2c) + \
                                        Y1s*(-Y2s + \
                                             sign_G*sign_psi*curvature))*\
                                       (iota_N0*Y1c - d_Y1s_d_varphi)) \
                                     + X2s*(iota_N0*Y1c*Y1c*\
                                            (sign_G*sign_psi - 2*X1c*Y1s) - \
                                            sign_G*sign_psi*Y1s*\
                                            (iota_N0*Y1s + \
                                             d_Y1c_d_varphi) + \
                                            Y1c*(-(sign_G*sign_psi) + \
                                                 2*X1c*Y1s)*\
                                            d_Y1s_d_varphi) + \
                                     sign_G*sign_psi*(X20*\
                                                      (Y1c*d_Y1c_d_varphi + \
                                                       Y1s*d_Y1s_d_varphi) + \
                                                      X2c*(-(Y1c*\
                                                             (2*iota_N0*Y1s + \
                                                              d_Y1c_d_varphi)) + \
                                                           Y1s*d_Y1s_d_varphi)) + \
                                     X1c*(-(X2c*\
                                            (Y1c*Y1c - Y1s*Y1s)*\
                                            (iota_N0*Y1c - \
                                             d_Y1s_d_varphi)) + \
                                          X20*(Y1c*Y1c + Y1s*Y1s)*\
                                          (iota_N0*Y1c - d_Y1s_d_varphi) \
                                          + sign_G*sign_psi*(Y1c*\
                                                             (2*iota_N0*Y2s - \
                                                              d_Y20_d_varphi + \
                                                              d_Y2c_d_varphi) + \
                                                             Y1s*\
                                                             (-2*iota_N0*Y2c + \
                                                              d_Y2s_d_varphi)))))/(lp*sign_psi)

    # Element 123
    grad_grad_B_alt[:,0,1,2] =(2*X1c*(Y1c*\
                                      (G2 + I2*iota - 2*lp*sign_G*B20 + \
                                       2*lp*sign_G*B2c + \
                                       2*B0*sign_G*iota_N0*Z2s + \
                                       2*B0*lp*sign_G*X20*curvature - \
                                       2*B0*lp*sign_G*X2c*curvature - \
                                       B0*sign_G*d_Z20_d_varphi + \
                                       B0*sign_G*d_Z2c_d_varphi) + \
                                      sign_G*Y1s*(2*lp*B2s + \
                                                  B0*(-2*iota_N0*Z2c - \
                                                      2*lp*X2s*curvature + \
                                                      d_Z2s_d_varphi))))/(lp)

    # Element 131
    grad_grad_B_alt[:,0,2,0] =(B0*(-(lp*sign_G*sign_psi*iota_N0*Y1c*Y1c*\
                                     torsion) + \
                                   lp*iota_N0*X1c*X1c*X1c*Y1s*\
                                   torsion + \
                                   X1c*X1c*(lp*lp*Y1s*Y1s*\
                                            torsion*torsion + \
                                            iota_N0*Y1s*d_Y1c_d_varphi - \
                                            iota_N0*Y1c*d_Y1s_d_varphi) + \
                                   sign_G*sign_psi*Y1c*\
                                   (iota_N0*d_X1c_d_varphi + \
                                    2*lp*torsion*d_Y1s_d_varphi) + \
                                   X1c*Y1s*\
                                   (2*lp*lp*sign_G*sign_psi*curvature*curvature - \
                                    lp*lp*sign_G*sign_psi*torsion*torsion - \
                                    iota_N0*Y1c*d_X1c_d_varphi + \
                                    lp*torsion*\
                                    (Y1s*d_Y1c_d_varphi - \
                                     Y1c*d_Y1s_d_varphi)) - \
                                   Y1s*(Y1s*\
                                        (lp*sign_G*sign_psi*iota_N0*torsion + \
                                         d_X1c_d_varphi*d_X1c_d_varphi) + \
                                        sign_G*sign_psi*(2*lp*torsion*\
                                                         d_Y1c_d_varphi - \
                                                         d2_X1c_d_varphi2))))/(lp*lp*sign_G)

    # Element 132
    grad_grad_B_alt[:,0,2,1] =(B0*(-(iota_N0*Y1c*Y1c*Y1s*\
                                     d_X1c_d_varphi) + \
                                   lp*X1c*X1c*Y1s*torsion*\
                                   (iota_N0*Y1c - d_Y1s_d_varphi) + \
                                   X1c*(-(iota_N0*Y1c*Y1c*\
                                          d_Y1s_d_varphi) + \
                                        Y1c*(lp*sign_G*sign_psi*iota_N0*\
                                             torsion + \
                                             iota_N0*Y1s*\
                                             d_Y1c_d_varphi + \
                                             d_Y1s_d_varphi*d_Y1s_d_varphi) - \
                                        Y1s*(lp*Y1s*torsion*\
                                             d_X1c_d_varphi + \
                                             d_Y1c_d_varphi*\
                                             d_Y1s_d_varphi - \
                                             lp*sign_G*sign_psi*d_torsion_d_varphi)) + \
                                   Y1s*(-(iota_N0*Y1s*Y1s*\
                                          d_X1c_d_varphi) - \
                                        Y1s*d_X1c_d_varphi*\
                                        d_Y1c_d_varphi + \
                                        sign_G*sign_psi*(2*lp*torsion*\
                                                         d_X1c_d_varphi + \
                                                         iota_N0*d_Y1s_d_varphi + \
                                                         d2_Y1c_d_varphi2)) + \
                                   Y1c*(sign_G*sign_psi*iota_N0*\
                                        d_Y1c_d_varphi + \
                                        Y1s*d_X1c_d_varphi*\
                                        d_Y1s_d_varphi - \
                                        sign_G*sign_psi*d2_Y1s_d_varphi2)))/\
                                        (lp*lp*sign_G)

    # Element 133
    grad_grad_B_alt[:,0,2,2] =-((B0*(Y1s*curvature*\
                                     d_X1c_d_varphi + \
                                     X1c*(iota_N0*Y1c*\
                                          curvature - \
                                          Y1s*d_curvature_d_varphi)))/\
                                (lp*sign_psi))

    # Element 211
    grad_grad_B_alt[:,1,0,0] =(-2*B0*X1c*(2*sign_G*sign_psi*iota_N0*X2c*\
                                          Y1s + iota_N0*X1c*X1c*\
                                          (Y1c*(Y20 - Y2c) + \
                                           sign_G*sign_psi*Y1s*curvature) - \
                                          X20*Y1c*Y1s*\
                                          d_X1c_d_varphi + \
                                          X2c*Y1c*Y1s*\
                                          d_X1c_d_varphi - \
                                          sign_G*sign_psi*Y20*d_X1c_d_varphi + \
                                          sign_G*sign_psi*Y2c*d_X1c_d_varphi + \
                                          X2s*(Y1c*\
                                               (-2*sign_G*sign_psi*iota_N0 + \
                                                iota_N0*X1c*Y1s) + \
                                               Y1s*Y1s*d_X1c_d_varphi) + \
                                          X1c*(-(iota_N0*X20*\
                                                 Y1c*Y1c) + \
                                               iota_N0*X2c*Y1c*Y1c - \
                                               sign_G*sign_psi*iota_N0*Y2s + \
                                               lp*sign_G*sign_psi*Y1s*Y1s*curvature*\
                                               torsion + \
                                               Y1s*Y20*\
                                               d_X1c_d_varphi - \
                                               Y1s*Y2c*\
                                               d_X1c_d_varphi) + \
                                          sign_G*sign_psi*Y1c*d_X20_d_varphi - \
                                          sign_G*sign_psi*Y1c*d_X2c_d_varphi - \
                                          sign_G*sign_psi*Y1s*d_X2s_d_varphi))/\
                                          (lp*sign_psi)

    # Element 212
    grad_grad_B_alt[:,1,0,1] =(-2*B0*X1c*(iota_N0*X2s*\
                                          Y1c*Y1c*Y1s + \
                                          iota_N0*X2s*Y1s*Y1s*Y1s + \
                                          iota_N0*X1c*Y1c*Y1c*\
                                          Y20 - sign_G*sign_psi*iota_N0*Y1s*\
                                          Y20 + iota_N0*X1c*\
                                          Y1s*Y1s*Y20 - \
                                          iota_N0*X1c*Y1c*Y1c*\
                                          Y2c + 3*sign_G*sign_psi*iota_N0*Y1s*\
                                          Y2c - iota_N0*X1c*\
                                          Y1s*Y1s*Y2c - \
                                          3*sign_G*sign_psi*iota_N0*Y1c*Y2s + \
                                          sign_G*sign_psi*iota_N0*X1c*Y1c*\
                                          Y1s*curvature - \
                                          lp*sign_G*sign_psi*X2s*Y1s*\
                                          torsion + \
                                          lp*X1c*X2s*Y1s*Y1s*\
                                          torsion - \
                                          lp*sign_G*sign_psi*X1c*Y20*\
                                          torsion + \
                                          lp*X1c*X1c*Y1s*Y20*\
                                          torsion + \
                                          lp*sign_G*sign_psi*X1c*Y2c*\
                                          torsion - \
                                          lp*X1c*X1c*Y1s*Y2c*\
                                          torsion + \
                                          X2s*Y1s*Y1s*\
                                          d_Y1c_d_varphi - \
                                          sign_G*sign_psi*Y20*d_Y1c_d_varphi + \
                                          X1c*Y1s*Y20*\
                                          d_Y1c_d_varphi + \
                                          sign_G*sign_psi*Y2c*d_Y1c_d_varphi - \
                                          X1c*Y1s*Y2c*\
                                          d_Y1c_d_varphi - \
                                          X2s*Y1c*Y1s*\
                                          d_Y1s_d_varphi - \
                                          X1c*Y1c*Y20*\
                                          d_Y1s_d_varphi + \
                                          X1c*Y1c*Y2c*\
                                          d_Y1s_d_varphi + \
                                          sign_G*sign_psi*Y2s*d_Y1s_d_varphi - \
                                          sign_G*sign_psi*X1c*Y1s*\
                                          curvature*d_Y1s_d_varphi - \
                                          X20*Y1c*\
                                          (iota_N0*Y1c*Y1c + iota_N0*Y1s*Y1s - \
                                           lp*sign_G*sign_psi*torsion + \
                                           Y1s*(lp*X1c*torsion + \
                                                d_Y1c_d_varphi) - \
                                           Y1c*d_Y1s_d_varphi) + \
                                          X2c*Y1c*\
                                          (iota_N0*Y1c*Y1c + iota_N0*Y1s*Y1s - \
                                           lp*sign_G*sign_psi*torsion + \
                                           Y1s*(lp*X1c*torsion + \
                                                d_Y1c_d_varphi) - \
                                           Y1c*d_Y1s_d_varphi) + \
                                          sign_G*sign_psi*Y1c*d_Y20_d_varphi - \
                                          sign_G*sign_psi*Y1c*d_Y2c_d_varphi - \
                                          sign_G*sign_psi*Y1s*d_Y2s_d_varphi))/\
                                          (lp*sign_psi)

    # Element 213
    grad_grad_B_alt[:,1,0,2] =(2*X1c*(2*lp*sign_G*sign_psi*B2s*\
                                      Y1s + Y1c*\
                                      (G2*sign_psi + I2*sign_psi*iota - \
                                       2*lp*sign_G*sign_psi*B20 + \
                                       2*lp*sign_G*sign_psi*B2c + \
                                       2*B0*sign_G*sign_psi*iota_N0*Z2s + \
                                       B0*lp*sign_G*sign_psi*X20*curvature - \
                                       B0*lp*sign_G*sign_psi*X2c*curvature + \
                                       B0*lp*X1c*X20*Y1s*\
                                       curvature - \
                                       B0*lp*X1c*X2c*Y1s*\
                                       curvature - \
                                       B0*sign_G*sign_psi*d_Z20_d_varphi + \
                                       B0*sign_G*sign_psi*d_Z2c_d_varphi) - \
                                      B0*(lp*X1c*X2s*Y1s*Y1s*\
                                          curvature + \
                                          lp*sign_G*sign_psi*X1c*\
                                          (-Y20 + Y2c)*\
                                          curvature + \
                                          Y1s*(2*sign_G*sign_psi*iota_N0*Z2c + \
                                               lp*sign_G*sign_psi*X2s*curvature + \
                                               lp*X1c*X1c*Y20*\
                                               curvature - \
                                               lp*X1c*X1c*Y2c*\
                                               curvature - \
                                               sign_G*sign_psi*d_Z2s_d_varphi))))/\
                                               (lp*sign_psi)

    # Element 221
    grad_grad_B_alt[:,1,1,0] =(2*B0*X1c*(iota_N0*X1c*X1c*X1c*\
                                         (Y20 - Y2c) + \
                                         X1c*X1c*(-(iota_N0*X20*\
                                                    Y1c) + \
                                                  iota_N0*X2c*Y1c + \
                                                  Y1s*(iota_N0*X2s + \
                                                       lp*(Y20 - Y2c)*\
                                                       torsion)) - \
                                         sign_G*sign_psi*(lp*X2s*Y1s*\
                                                          torsion + \
                                                          X2c*(lp*Y1c*\
                                                               torsion - d_X1c_d_varphi) + \
                                                          X20*(-(lp*Y1c*\
                                                                 torsion) + d_X1c_d_varphi)) \
                                         + X1c*(-(lp*X20*Y1c*\
                                                  Y1s*torsion) + \
                                                lp*X2c*Y1c*Y1s*\
                                                torsion - \
                                                lp*sign_G*sign_psi*Y20*torsion + \
                                                lp*sign_G*sign_psi*Y2c*torsion + \
                                                X2s*(-3*sign_G*sign_psi*iota_N0 + \
                                                     lp*Y1s*Y1s*torsion) + \
                                                sign_G*sign_psi*d_X20_d_varphi - \
                                                sign_G*sign_psi*d_X2c_d_varphi)))/\
                                                (lp*sign_psi)

    # Element 222
    grad_grad_B_alt[:,1,1,1] =(-2*B0*X1c*(sign_G*sign_psi*\
                                          (X20 - X2c)*\
                                          (iota_N0*Y1s + d_Y1c_d_varphi) + \
                                          X2s*(sign_G*sign_psi - \
                                               X1c*Y1s)*\
                                          (iota_N0*Y1c - d_Y1s_d_varphi) - \
                                          X1c*X1c*(Y20 - Y2c)*\
                                          (iota_N0*Y1c - d_Y1s_d_varphi) + \
                                          X1c*(X20*Y1c*\
                                               (iota_N0*Y1c - d_Y1s_d_varphi) \
                                               + X2c*Y1c*\
                                               (-(iota_N0*Y1c) + \
                                                d_Y1s_d_varphi) + \
                                               sign_G*sign_psi*(2*iota_N0*Y2s - \
                                                                d_Y20_d_varphi + \
                                                                d_Y2c_d_varphi))))/(lp*sign_psi)

    # Element 223
    grad_grad_B_alt[:,1,1,2] =(-2*X1c*X1c*(G2 + I2*iota - 2*lp*sign_G*B20 + \
                                           2*lp*sign_G*B2c + 2*B0*sign_G*iota_N0*Z2s + \
                                           2*B0*lp*sign_G*X20*curvature - \
                                           2*B0*lp*sign_G*X2c*curvature - \
                                           B0*sign_G*d_Z20_d_varphi + \
                                           B0*sign_G*d_Z2c_d_varphi))/(lp)

    # Element 231
    grad_grad_B_alt[:,1,2,0] =(B0*X1c*(lp*iota_N0*Y1c*\
                                       (sign_G*sign_psi + X1c*Y1s)*\
                                       torsion + \
                                       (-(sign_G*sign_psi*iota_N0) + \
                                        lp*Y1s*Y1s*torsion)*\
                                       d_X1c_d_varphi + \
                                       iota_N0*X1c*X1c*d_Y1s_d_varphi - \
                                       2*lp*sign_G*sign_psi*torsion*\
                                       d_Y1s_d_varphi + \
                                       lp*X1c*Y1s*torsion*\
                                       d_Y1s_d_varphi - \
                                       lp*sign_G*sign_psi*Y1s*d_torsion_d_varphi))/\
                                       (lp*lp*sign_G)

    # Element 232
    grad_grad_B_alt[:,1,2,1] =(B0*X1c*(lp*iota_N0*Y1c*Y1c*\
                                       Y1s*torsion + \
                                       lp*iota_N0*Y1s*Y1s*Y1s*torsion - \
                                       lp*lp*sign_G*sign_psi*Y1s*torsion*torsion - \
                                       sign_G*sign_psi*iota_N0*d_Y1c_d_varphi + \
                                       lp*Y1s*Y1s*torsion*\
                                       d_Y1c_d_varphi - \
                                       lp*Y1c*Y1s*torsion*\
                                       d_Y1s_d_varphi + \
                                       X1c*(-(lp*sign_G*sign_psi*iota_N0*\
                                              torsion) + \
                                            lp*lp*Y1s*Y1s*torsion*torsion + \
                                            (iota_N0*Y1c - d_Y1s_d_varphi)*\
                                            d_Y1s_d_varphi) + \
                                       sign_G*sign_psi*d2_Y1s_d_varphi2))/(lp*lp*sign_G)

    # Element 233
    grad_grad_B_alt[:,1,2,2] =(B0*X1c*curvature*\
                               (iota_N0*X1c + \
                                2*lp*Y1s*torsion))/(lp*sign_psi)

    # Element 311
    grad_grad_B_alt[:,2,0,0] =(B0*(2*lp*lp*sign_G*sign_psi*curvature*curvature + \
                                   lp*iota_N0*X1c*X1c*torsion - \
                                   lp*iota_N0*Y1c*Y1c*torsion - \
                                   lp*iota_N0*Y1s*Y1s*torsion + \
                                   iota_N0*Y1c*d_X1c_d_varphi + \
                                   iota_N0*X1c*d_Y1c_d_varphi - \
                                   lp*Y1s*torsion*\
                                   d_Y1c_d_varphi + \
                                   lp*Y1c*torsion*\
                                   d_Y1s_d_varphi + \
                                   d_X1c_d_varphi*d_Y1s_d_varphi + \
                                   Y1s*d2_X1c_d_varphi2))/\
                                   (lp*lp*sign_psi)

    # Element 312
    grad_grad_B_alt[:,2,0,1] =(B0*(lp*X1c*(2*iota_N0*Y1c*\
                                           torsion + \
                                           Y1s*d_torsion_d_varphi) + \
                                   Y1s*(2*lp*torsion*\
                                        d_X1c_d_varphi + \
                                        2*iota_N0*d_Y1s_d_varphi + \
                                        d2_Y1c_d_varphi2) + \
                                   Y1c*(2*iota_N0*d_Y1c_d_varphi - \
                                        d2_Y1s_d_varphi2)))/(lp*lp*sign_psi)

    # Element 313
    grad_grad_B_alt[:,2,0,2] =(B0*(-(iota_N0*X1c*Y1c*\
                                     curvature) - \
                                   Y1s*curvature*\
                                   d_X1c_d_varphi + \
                                   sign_G*sign_psi*d_curvature_d_varphi))/(lp*sign_psi)

    # Element 321
    grad_grad_B_alt[:,2,1,0] =(B0*X1c*(2*lp*iota_N0*Y1c*\
                                       torsion - \
                                       2*iota_N0*d_X1c_d_varphi - \
                                       lp*(2*torsion*d_Y1s_d_varphi + \
                                           Y1s*d_torsion_d_varphi)))/\
                                           (lp*lp*sign_psi)

    # Element 322
    grad_grad_B_alt[:,2,1,1] =-((B0*(iota_N0*Y1c*d_X1c_d_varphi + \
                                     iota_N0*X1c*d_Y1c_d_varphi - \
                                     d_X1c_d_varphi*\
                                     d_Y1s_d_varphi + \
                                     lp*torsion*\
                                     (iota_N0*X1c*X1c - \
                                      iota_N0*Y1c*Y1c - \
                                      Y1s*(iota_N0*Y1s + \
                                           d_Y1c_d_varphi) + \
                                      Y1c*d_Y1s_d_varphi) - \
                                     X1c*d2_Y1s_d_varphi2))/\
                                (lp*lp*sign_psi))

    # Element 323
    grad_grad_B_alt[:,2,1,2] =(B0*curvature*(iota_N0*X1c*X1c + \
                                             lp*sign_G*sign_psi*torsion + \
                                             lp*X1c*Y1s*torsion))/\
                                             (lp*sign_psi)

    # Element 331
    grad_grad_B_alt[:,2,2,0] =(B0*(-(iota_N0*X1c*Y1c*\
                                     curvature) - \
                                   Y1s*curvature*\
                                   d_X1c_d_varphi + \
                                   sign_G*sign_psi*d_curvature_d_varphi))/(lp*sign_psi)

    # Element 332
    grad_grad_B_alt[:,2,2,1] =-((B0*curvature*(iota_N0*Y1c*Y1c + \
                                               iota_N0*Y1s*Y1s - \
                                               lp*sign_G*sign_psi*torsion + \
                                               Y1s*(lp*X1c*torsion + \
                                                    d_Y1c_d_varphi) - \
                                               Y1c*d_Y1s_d_varphi))/\
                                (lp*sign_psi))

    # Element 333
    grad_grad_B_alt[:,2,2,2] =(-2*B0*curvature*curvature)/sign_G

    self.grad_grad_B_alt = grad_grad_B_alt


def Bfield_cylindrical(self, r=0, theta=0):
    '''
    Function to calculate the magnetic field vector B=(B_R,B_phi,B_Z) at
    every point along the axis (hence with nphi points) where R, phi and Z
    are the standard cylindrical coordinates for a given
    near-axis radius r and a Boozer poloidal angle vartheta (not theta).
    The formulae implemented here are eq (3.5) and (3.6) of
    Landreman (2021): Figures of merit for stellarators near the magnetic axis, JPP

    Args:
      r: the near-axis radius
      theta: the Boozer poloidal angle vartheta (= theta-N*phi)
    '''
    
    # Define auxiliary variables
    t = self.tangent_cylindrical.transpose()
    n = self.normal_cylindrical.transpose()
    b = self.binormal_cylindrical.transpose()
    B0 = self.B0
    sG = self.sG
    G0 = self.G0
    X1c = self.X1c
    X1s = self.X1s
    Y1c = self.Y1c
    Y1s = self.Y1s
    d_l_d_varphi = self.d_l_d_varphi
    curvature = self.curvature
    torsion = self.torsion
    iotaN = self.iotaN
    d_X1c_d_varphi = self.d_X1c_d_varphi
    d_X1s_d_varphi = self.d_X1s_d_varphi
    d_Y1s_d_varphi = self.d_Y1s_d_varphi
    d_Y1c_d_varphi = self.d_Y1c_d_varphi

    B0_vector = sG * B0 * t

    if r == 0:
        return B0_vector
    else:
        factor = B0 * B0 / G0
        B1_vector_t = factor * (X1c * np.cos(theta) + X1s * np.sin(theta)) * d_l_d_varphi * curvature
        B1_vector_n = factor * (np.cos(theta) * (d_X1c_d_varphi - Y1c * d_l_d_varphi * torsion + iotaN * X1s) \
                                + np.sin(theta) * (d_X1s_d_varphi - Y1s * d_l_d_varphi * torsion - iotaN * X1c))
        B1_vector_b = factor * (np.cos(theta) * (d_Y1c_d_varphi + X1c * d_l_d_varphi * torsion + iotaN * Y1s) \
                                + np.sin(theta) * (d_Y1s_d_varphi + X1s * d_l_d_varphi * torsion - iotaN * Y1c))

        B1_vector = B1_vector_t * t + B1_vector_n * n + B1_vector_b * b
        B_vector_cylindrical = B0_vector + r * B1_vector

        return B_vector_cylindrical

def Bfield_cartesian(self, r=0, theta=0):
    '''
    Function to calculate the magnetic field vector B=(B_x,B_y,B_z) at
    every point along the axis (hence with nphi points) where x, y and z
    are the standard cartesian coordinates for a given
    near-axis radius r and a Boozer poloidal angle vartheta (not theta).

    Args:
      r: the near-axis radius
      theta: the Boozer poloidal angle vartheta (= theta-N*phi)
    '''
    B_vector_cylindrical = self.Bfield_cylindrical(r,theta)
    phi = self.phi

    B_x = np.cos(phi) * B_vector_cylindrical[0] - np.sin(phi) * B_vector_cylindrical[1]
    B_y = np.sin(phi) * B_vector_cylindrical[0] + np.cos(phi) * B_vector_cylindrical[1]
    B_z = B_vector_cylindrical[2]

    B_vector_cartesian = np.array([B_x, B_y, B_z])

    return B_vector_cartesian

def grad_B_tensor_cartesian(self):
    '''
    Function to calculate the gradient of the magnetic field vector B=(B_x,B_y,B_z)
    at every point along the axis (hence with nphi points) where x, y and z
    are the standard cartesian coordinates.
    '''

    B0, B1, B2 = self.Bfield_cylindrical()
    nablaB = self.grad_B_tensor_cylindrical
    cosphi = np.cos(self.phi)
    sinphi = np.sin(self.phi)
    R0 = self.R0

    grad_B_vector_cartesian = np.array([
[cosphi**2*nablaB[0, 0] - cosphi*sinphi*(nablaB[0, 1] + nablaB[1, 0]) + 
   sinphi**2*nablaB[1, 1], cosphi**2*nablaB[0, 1] - sinphi**2*nablaB[1, 0] + 
   cosphi*sinphi*(nablaB[0, 0] - nablaB[1, 1]), cosphi*nablaB[0, 2] - 
   sinphi*nablaB[1, 2]], [-(sinphi**2*nablaB[0, 1]) + cosphi**2*nablaB[1, 0] + 
   cosphi*sinphi*(nablaB[0, 0] - nablaB[1, 1]), sinphi**2*nablaB[0, 0] + 
   cosphi*sinphi*(nablaB[0, 1] + nablaB[1, 0]) + cosphi**2*nablaB[1, 1], 
  sinphi*nablaB[0, 2] + cosphi*nablaB[1, 2]], 
 [cosphi*nablaB[2, 0] - sinphi*nablaB[2, 1], sinphi*nablaB[2, 0] + cosphi*nablaB[2, 1], 
  nablaB[2, 2]]
    ])

    return grad_B_vector_cartesian

def grad_grad_B_tensor_cylindrical(self):
    '''
    Function to calculate the gradient of of the gradient the magnetic field
    vector B=(B_R,B_phi,B_Z) at every point along the axis (hence with nphi points)
    where R, phi and Z are the standard cylindrical coordinates.
    '''
    return np.transpose(self.grad_grad_B,(1,2,3,0))

def grad_grad_B_tensor_cartesian(self):
    '''
    Function to calculate the gradient of of the gradient the magnetic field
    vector B=(B_x,B_y,B_z) at every point along the axis (hence with nphi points)
    where x, y and z are the standard cartesian coordinates.
    '''
    nablanablaB = self.grad_grad_B_tensor_cylindrical()
    cosphi = np.cos(self.phi)
    sinphi = np.sin(self.phi)

    grad_grad_B_vector_cartesian = np.array([[
[cosphi**3*nablanablaB[0, 0, 0] - cosphi**2*sinphi*(nablanablaB[0, 0, 1] + 
      nablanablaB[0, 1, 0] + nablanablaB[1, 0, 0]) + 
    cosphi*sinphi**2*(nablanablaB[0, 1, 1] + nablanablaB[1, 0, 1] + 
      nablanablaB[1, 1, 0]) - sinphi**3*nablanablaB[1, 1, 1], 
   cosphi**3*nablanablaB[0, 0, 1] + cosphi**2*sinphi*(nablanablaB[0, 0, 0] - 
      nablanablaB[0, 1, 1] - nablanablaB[1, 0, 1]) + sinphi**3*nablanablaB[1, 1, 0] - 
    cosphi*sinphi**2*(nablanablaB[0, 1, 0] + nablanablaB[1, 0, 0] - 
      nablanablaB[1, 1, 1]), cosphi**2*nablanablaB[0, 0, 2] - 
    cosphi*sinphi*(nablanablaB[0, 1, 2] + nablanablaB[1, 0, 2]) + 
    sinphi**2*nablanablaB[1, 1, 2]], [cosphi**3*nablanablaB[0, 1, 0] + 
    sinphi**3*nablanablaB[1, 0, 1] + cosphi**2*sinphi*(nablanablaB[0, 0, 0] - 
      nablanablaB[0, 1, 1] - nablanablaB[1, 1, 0]) - 
    cosphi*sinphi**2*(nablanablaB[0, 0, 1] + nablanablaB[1, 0, 0] - 
      nablanablaB[1, 1, 1]), cosphi**3*nablanablaB[0, 1, 1] - 
    sinphi**3*nablanablaB[1, 0, 0] + cosphi*sinphi**2*(nablanablaB[0, 0, 0] - 
      nablanablaB[1, 0, 1] - nablanablaB[1, 1, 0]) + 
    cosphi**2*sinphi*(nablanablaB[0, 0, 1] + nablanablaB[0, 1, 0] - 
      nablanablaB[1, 1, 1]), cosphi**2*nablanablaB[0, 1, 2] - 
    sinphi**2*nablanablaB[1, 0, 2] + cosphi*sinphi*(nablanablaB[0, 0, 2] - 
      nablanablaB[1, 1, 2])], [cosphi**2*nablanablaB[0, 2, 0] - 
    cosphi*sinphi*(nablanablaB[0, 2, 1] + nablanablaB[1, 2, 0]) + 
    sinphi**2*nablanablaB[1, 2, 1], cosphi**2*nablanablaB[0, 2, 1] - 
    sinphi**2*nablanablaB[1, 2, 0] + cosphi*sinphi*(nablanablaB[0, 2, 0] - 
      nablanablaB[1, 2, 1]), cosphi*nablanablaB[0, 2, 2] - 
    sinphi*nablanablaB[1, 2, 2]]], 
 [[sinphi**3*nablanablaB[0, 1, 1] + cosphi**3*nablanablaB[1, 0, 0] + 
    cosphi**2*sinphi*(nablanablaB[0, 0, 0] - nablanablaB[1, 0, 1] - 
      nablanablaB[1, 1, 0]) - cosphi*sinphi**2*(nablanablaB[0, 0, 1] + 
      nablanablaB[0, 1, 0] - nablanablaB[1, 1, 1]), -(sinphi**3*nablanablaB[0, 1, 0]) + 
    cosphi**3*nablanablaB[1, 0, 1] + cosphi*sinphi**2*(nablanablaB[0, 0, 0] - 
      nablanablaB[0, 1, 1] - nablanablaB[1, 1, 0]) + 
    cosphi**2*sinphi*(nablanablaB[0, 0, 1] + nablanablaB[1, 0, 0] - 
      nablanablaB[1, 1, 1]), -(sinphi**2*nablanablaB[0, 1, 2]) + 
    cosphi**2*nablanablaB[1, 0, 2] + cosphi*sinphi*(nablanablaB[0, 0, 2] - 
      nablanablaB[1, 1, 2])], [-(sinphi**3*nablanablaB[0, 0, 1]) + 
    cosphi*sinphi**2*(nablanablaB[0, 0, 0] - nablanablaB[0, 1, 1] - 
      nablanablaB[1, 0, 1]) + cosphi**3*nablanablaB[1, 1, 0] + 
    cosphi**2*sinphi*(nablanablaB[0, 1, 0] + nablanablaB[1, 0, 0] - 
      nablanablaB[1, 1, 1]), sinphi**3*nablanablaB[0, 0, 0] + 
    cosphi*sinphi**2*(nablanablaB[0, 0, 1] + nablanablaB[0, 1, 0] + 
      nablanablaB[1, 0, 0]) + cosphi**2*sinphi*(nablanablaB[0, 1, 1] + 
      nablanablaB[1, 0, 1] + nablanablaB[1, 1, 0]) + cosphi**3*nablanablaB[1, 1, 1], 
   sinphi**2*nablanablaB[0, 0, 2] + cosphi*sinphi*(nablanablaB[0, 1, 2] + 
      nablanablaB[1, 0, 2]) + cosphi**2*nablanablaB[1, 1, 2]], 
  [-(sinphi**2*nablanablaB[0, 2, 1]) + cosphi**2*nablanablaB[1, 2, 0] + 
    cosphi*sinphi*(nablanablaB[0, 2, 0] - nablanablaB[1, 2, 1]), 
   sinphi**2*nablanablaB[0, 2, 0] + cosphi*sinphi*(nablanablaB[0, 2, 1] + 
      nablanablaB[1, 2, 0]) + cosphi**2*nablanablaB[1, 2, 1], 
   sinphi*nablanablaB[0, 2, 2] + cosphi*nablanablaB[1, 2, 2]]], 
 [[cosphi**2*nablanablaB[2, 0, 0] - cosphi*sinphi*(nablanablaB[2, 0, 1] + 
      nablanablaB[2, 1, 0]) + sinphi**2*nablanablaB[2, 1, 1], 
   cosphi**2*nablanablaB[2, 0, 1] - sinphi**2*nablanablaB[2, 1, 0] + 
    cosphi*sinphi*(nablanablaB[2, 0, 0] - nablanablaB[2, 1, 1]), 
   cosphi*nablanablaB[2, 0, 2] - sinphi*nablanablaB[2, 1, 2]], 
  [-(sinphi**2*nablanablaB[2, 0, 1]) + cosphi**2*nablanablaB[2, 1, 0] + 
    cosphi*sinphi*(nablanablaB[2, 0, 0] - nablanablaB[2, 1, 1]), 
   sinphi**2*nablanablaB[2, 0, 0] + cosphi*sinphi*(nablanablaB[2, 0, 1] + 
      nablanablaB[2, 1, 0]) + cosphi**2*nablanablaB[2, 1, 1], 
   sinphi*nablanablaB[2, 0, 2] + cosphi*nablanablaB[2, 1, 2]], 
  [cosphi*nablanablaB[2, 2, 0] - sinphi*nablanablaB[2, 2, 1], 
   sinphi*nablanablaB[2, 2, 0] + cosphi*nablanablaB[2, 2, 1], nablanablaB[2, 2, 2]]
      ]])

    return grad_grad_B_vector_cartesian



"""
This module contains the routines to output a
near-axis boundary to a VMEC input file
"""
from datetime import datetime
import numpy as np
from .Frenet_to_cylindrical import Frenet_to_cylindrical
from .util import mu0, to_Fourier

def to_vmec(self, filename, r=0.1, params=dict(), ntheta=20, ntorMax=14):
    """
    Outputs the near-axis configuration calculated with pyQIC to
    a text file that is able to be read by VMEC.

    Args:
        filename: name of the text file to be created
        r:  near-axis radius r of the desired boundary surface
        params: a Python dict() instance containing one/several of the following parameters: mpol,
          delt, nstep, tcon0, ns_array, ftol_array, niter_array
        ntheta: resolution in the poloidal angle theta for the Frenet_to_cylindrical and VMEC calculations
        ntorMax: maximum number of NTOR in the resulting VMEC input file
    """
    if "mpol" not in params.keys():
        mpol1d = 100 # maximum number of mode numbers VMEC can handle
        mpol = int(np.floor(min(ntheta / 2, mpol1d)))
    else:
        mpol = int(params["mpol"])
    if "ntor" not in params.keys():
        # We should be able to resolve (N_phi-1)/2 modes (note integer division!), but in case N_phi is very large, don't attempt more than the vmec arrays can handle.
        ntord = 100 # maximum number of mode numbers VMEC can handle
        ntor = int(min(self.nphi / 2, ntord))
    else:
        ntor = int(params["ntor"])
    if "delt" not in params.keys():
        params["delt"] = 0.9
    if "nstep" not in params.keys():
        params["nstep"] = 200
    if "tcon0" not in params.keys():
        params["tcon0"] = 2.0
    if "ns_array" not in params.keys():
        params["ns_array"] = [16,49,101]
    if "ftol_array" not in params.keys():
        params["ftol_array"] = [1e-14,1e-13,1e-13]
    if "niter_array" not in params.keys():
        params["niter_array"] = [2000,2000,2000]

    phiedge = np.pi * r * r * self.spsi * self.Bbar

    # Set pressure Profile
    temp = - self.p2 * r * r
    am = [temp,-temp]
    pmass_type='power_series'
    pres_scale=1

    # Set current profile:
    ncurr = 1
    pcurr_type = 'power_series'
    ac = [1]
    curtor = 2 * np.pi / mu0 * self.I2 * r * r

    # Get surface shape at fixed off-axis toroidal angle phi
    R_2D, Z_2D, phi0_2D = self.Frenet_to_cylindrical(r, ntheta)
    
    # Fourier transform the result.
    # This is not a rate-limiting step, so for clarity of code, we don't bother with an FFT.
    RBC, RBS, ZBC, ZBS = to_Fourier(R_2D, Z_2D, self.nfp, mpol, ntor, self.lasym)

    # Write to VMEC file
    file_object = open(filename,"w+")
    file_object.write("! This &INDATA namelist was generated by pyQIC: github.com/rogeriojorge/pyQIC\n")
    file_object.write("! Date: "+datetime.now().strftime("%B %d, %Y")+", Time: "+datetime.now().strftime("%H:%M:%S")+" UTC"+datetime.now().astimezone().strftime("%z")+"\n")
    file_object.write('! Near-axis parameters:  radius r = '+str(r)+', etabar = '+str(self.etabar)+'\n')
    file_object.write('! nphi = '+str(self.nphi)+', order = '+self.order+', sigma0 = '+str(self.sigma0)+', I2 = '+str(self.I2)+', B0 = '+str(np.mean(self.Bbar))+'\n')
    file_object.write('! Resolution parameters: ntheta = '+str(ntheta)+', mpol = '+str(mpol)+', ntor = '+str(ntor)+'\n')
    file_object.write('!----- Runtime Parameters -----\n')
    file_object.write('&INDATA\n')
    file_object.write('  DELT = '+str(params["delt"])+'\n')
    file_object.write('  NSTEP = '+str(params["nstep"])+'\n')
    file_object.write('  TCON0 = '+str(params["tcon0"])+'\n')
    file_object.write('  NS_ARRAY = '+str(params["ns_array"])[1:-1]+'\n')
    file_object.write('  FTOL_ARRAY = '+str(params["ftol_array"])[1:-1]+'\n')
    file_object.write('  NITER_ARRAY = '+str(params["niter_array"])[1:-1]+'\n')
    file_object.write('!----- Grid Parameters -----\n')
    file_object.write('  LASYM = '+str(self.lasym)+'\n')
    file_object.write('  NFP = '+str(self.nfp)+'\n')
    file_object.write('  MPOL = '+str(mpol)+'\n')
    file_object.write('  NTOR = '+str(min(ntor,ntorMax))+'\n')
    file_object.write('  PHIEDGE = '+str(phiedge)+'\n')
    file_object.write('!----- Pressure Parameters -----\n')
    file_object.write('  PRES_SCALE = '+str(pres_scale)+'\n')
    file_object.write("  PMASS_TYPE = '"+pmass_type+"'\n")
    file_object.write('  AM = '+str(am)[1:-1]+'\n')
    file_object.write('!----- Free Boundary Parameters -----\n')
    file_object.write('  LFREEB = F')
    file_object.write('!----- Current/Iota Parameters -----\n')
    file_object.write('  CURTOR = '+str(curtor)+'\n')
    file_object.write('  NCURR = '+str(ncurr)+'\n')
    file_object.write("  PCURR_TYPE = '"+pcurr_type+"'\n")
    file_object.write('  AC = '+str(ac)[1:-1]+'\n')
    file_object.write('!----- Axis Parameters -----\n')
    # To convert sin(...) modes to vmec, we introduce a minus sign. This is because in vmec,
    # R and Z ~ sin(m theta - n phi), which for m=0 is sin(-n phi) = -sin(n phi).
    file_object.write('  RAXIS_CC = '+str(self.rc)[1:-1]+'\n')
    if self.lasym:
        file_object.write('  RAXIS_CS = '+str(-self.rs)[1:-1]+'\n')
        file_object.write('  ZAXIS_CC = '+str(self.zc)[1:-1]+'\n')
    file_object.write('  ZAXIS_CS = '+str(-self.zs)[1:-1]+'\n')
    file_object.write('!----- Boundary Parameters -----\n')
    for m in range(mpol+1):
        for n in range(-ntor,ntor+1):
            if RBC[n+ntor,m]!=0 or ZBS[n+ntor,m]!=0:
                file_object.write(    '  RBC('+f"{n:03d}"+','+f"{m:03d}"+') = '+f"{RBC[n+ntor,m]:+.16e}"+',    ZBS('+f"{n:03d}"+','+f"{m:03d}"+') = '+f"{ZBS[n+ntor,m]:+.16e}"+'\n')
                if self.lasym:
                    file_object.write('  RBS('+f"{n:03d}"+','+f"{m:03d}"+') = '+f"{RBS[n+ntor,m]:+.16e}"+',    ZBC('+f"{n:03d}"+','+f"{m:03d}"+') = '+f"{ZBC[n+ntor,m]:+.16e}"+'\n')
    file_object.write('/\n')
    file_object.close()

    self.RBC = RBC.transpose()
    self.ZBS = ZBS.transpose()
    if self.lasym:
        self.RBS = RBS.transpose()
        self.ZBC = ZBC.transpose()
    else:
        self.RBS = RBS
        self.ZBC = ZBC



#!/usr/bin/env python3

import unittest
import numpy as np
from qsc.newton import newton

class NewtonTests(unittest.TestCase):

    def test_2d(self):
        """
        Try a little 2D nonlinear example.
        """
        def func(x):
            return np.array([x[1] - np.exp(x[0]), x[0] + x[1]])
        
        def jac(x):
            return np.array([[-np.exp(x[0]), 1],
                             [1, 1]])
        x0 = np.zeros(2)
        soln = newton(func, x0, jac)
        np.testing.assert_allclose(soln, [-0.5671432904097838, 0.5671432904097838])
        
if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import unittest
import os
from scipy.io import netcdf
import numpy as np
import logging
from qsc.qsc import Qsc
from qsc.util import to_Fourier

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

def check_r2(s):
    """
    Verify that the O(r^2) equations have been solved, using a
    different method than the method used to solve them originally.
    """
    B0_over_abs_G0 = s.B0 / np.abs(s.G0)
    abs_G0_over_B0 = 1 / B0_over_abs_G0
    G0_over_Bbar = s.G0 / s.Bbar
    X1c = s.X1c
    X1s = s.X1s
    Y1s = s.Y1s
    Y1c = s.Y1c
    d_d_varphi = s.d_d_varphi
    iota_N = s.iotaN
    curvature = s.curvature
    torsion = s.torsion
    B0 = s.B0
    Bbar = s.Bbar
    sG = s.sG
    I2_over_Bbar = s.I2 / s.Bbar
    X20 = s.X20
    X2s = s.X2s
    X2c = s.X2c
    Y20 = s.Y20
    Y2s = s.Y2s
    Y2c = s.Y2c
    Z20 = s.Z20
    Z2s = s.Z2s
    Z2c = s.Z2c
    beta_0 = s.beta_0
    beta_1s = s.beta_1s
    beta_1c = s.beta_1c
    
    fX0 = np.matmul(d_d_varphi, X20) - torsion * abs_G0_over_B0 * Y20 + curvature * abs_G0_over_B0 * Z20 \
        -4*G0_over_Bbar*(Y2c * Z2s - Y2s * Z2c) \
        - I2_over_Bbar * (0.5*curvature * (X1s * Y1s + X1c * Y1c) - 2 * Y20) * abs_G0_over_B0 \
        - 0.5 * beta_0 * curvature * abs_G0_over_B0 * (X1s * Y1c - X1c * Y1s) - 0.5 * abs_G0_over_B0 * (beta_1c * Y1s - beta_1s * Y1c)

    fXs = np.matmul(d_d_varphi, X2s) - 2 * iota_N * X2c - torsion * abs_G0_over_B0 * Y2s + curvature * abs_G0_over_B0 * Z2s \
        -4*G0_over_Bbar*(-Y20 * Z2c + Y2c * Z20) \
        -I2_over_Bbar * (0.5*curvature * (X1s * Y1c + X1c * Y1s) - 2 * Y2s) * abs_G0_over_B0 \
        - beta_0 * abs_G0_over_B0 * (-2 * Y2c + 0.5 * curvature * (X1c * Y1c - X1s * Y1s)) - 0.5 * abs_G0_over_B0 * (beta_1s * Y1s - beta_1c * Y1c)

    fXc = np.matmul(d_d_varphi, X2c) + 2 * iota_N * X2s - torsion * abs_G0_over_B0 * Y2c + curvature * abs_G0_over_B0 * Z2c \
        -4*G0_over_Bbar*(Y20 * Z2s - Y2s * Z20) \
        -I2_over_Bbar * (0.5*curvature * (X1c * Y1c - X1s * Y1s) - 2 * Y2c) * abs_G0_over_B0 \
        - beta_0 * abs_G0_over_B0 * (2 * Y2s - 0.5 * curvature * (X1c * Y1s + X1s * Y1c)) - 0.5 * abs_G0_over_B0 * (beta_1c * Y1s + beta_1s * Y1c)

    fY0 = np.matmul(d_d_varphi, Y20) + torsion * abs_G0_over_B0 * X20 - 4*G0_over_Bbar*(X2s * Z2c - X2c * Z2s) \
        -I2_over_Bbar * (-0.5*curvature*(X1c*X1c + X1s*X1s) + 2*X20) * abs_G0_over_B0 - abs_G0_over_B0 * (beta_1s * X1c - beta_1c * X1s) / 2

    fYs = np.matmul(d_d_varphi, Y2s) - 2 * iota_N * Y2c + torsion * abs_G0_over_B0 * X2s \
        -4*G0_over_Bbar*(X20 * Z2c - X2c * Z20) -I2_over_Bbar*(-curvature * X1s * X1c + 2 * X2s)*abs_G0_over_B0\
        - beta_0 * abs_G0_over_B0 * (2*X2c + 0.5*curvature*(X1s*X1s - X1c*X1c)) - 0.5 * abs_G0_over_B0 * (beta_1c*X1c - beta_1s*X1s)

    fYc = np.matmul(d_d_varphi, Y2c) + 2 * iota_N * Y2s + torsion * abs_G0_over_B0 * X2c \
        -4*G0_over_Bbar*(X2s * Z20 - X20 * Z2s) \
        -I2_over_Bbar * (0.5*curvature * (X1s * X1s - X1c * X1c) + 2 * X2c) * abs_G0_over_B0 \
        -beta_0 * abs_G0_over_B0 * (-2 * X2s + curvature * X1s * X1c) + abs_G0_over_B0 * (beta_1c * X1s + beta_1s * X1c) / 2

    eq1residual = -X1s * fX0 + X1c * fXs - X1s * fXc - Y1s * fY0 + Y1c * fYs - Y1s * fYc
    eq2residual = -X1c * fX0 + X1s * fXs + X1c * fXc - Y1c * fY0 + Y1s * fYs + Y1c * fYc
    # Now check the two equations that were used to determine Y2s and Y2c:
    eq3residual = -X1s * Y2s - X1c * Y2c + X1c * Y20 + X2s * Y1s + X2c * Y1c - X20 * Y1c + 0.5 * sG * X1s * curvature * Bbar / B0
    eq4residual = -X1s * Y2c + X1c * Y2s - X1s * Y20 + X2c * Y1s - X2s * Y1c + X20 * Y1s + 0.5 * sG * X1c * curvature * Bbar / B0

    logger.info("max(abs(eq1residual)): {}".format(np.max(np.abs(eq1residual))))
    logger.info("max(abs(eq2residual)): {}".format(np.max(np.abs(eq2residual))))
    logger.info("max(abs(eq3residual)): {}".format(np.max(np.abs(eq3residual))))
    logger.info("max(abs(eq4residual)): {}".format(np.max(np.abs(eq4residual))))

    atol = 1e-8
    np.testing.assert_allclose(eq1residual, np.zeros(s.nphi), atol=atol)
    np.testing.assert_allclose(eq2residual, np.zeros(s.nphi), atol=atol)
    np.testing.assert_allclose(eq3residual, np.zeros(s.nphi), atol=atol)
    np.testing.assert_allclose(eq4residual, np.zeros(s.nphi), atol=atol)


def fortran_plot_single(filename, ntheta=150, nphi = 4):
    """
    Function to extract boundary arrays from the fortran files
    """
    abs_filename = os.path.join(os.path.dirname(__file__), filename)
    f = netcdf.netcdf_file(abs_filename,mode='r',mmap=False)
    r = f.variables['r'][()]
    nfp = f.variables['nfp'][()]
    nphi_axis = f.variables['N_phi'][()]
    mpol = f.variables['mpol'][()]
    ntor = f.variables['ntor'][()]
    RBC = f.variables['RBC'][()]
    RBS = f.variables['RBS'][()]
    ZBC = f.variables['ZBC'][()]
    ZBS = f.variables['ZBS'][()]
    R0c = f.variables['R0c'][()]
    R0s = f.variables['R0s'][()]
    Z0c = f.variables['Z0c'][()]
    Z0s = f.variables['Z0s'][()]

    theta1D = np.linspace(0,2*np.pi,ntheta)
    phi1D = np.linspace(0,2*np.pi,nphi)
    phi2D,theta2D = np.meshgrid(phi1D,theta1D)

    R = np.zeros((ntheta,nphi))
    z = np.zeros((ntheta,nphi))
    for m in range(mpol+1):
        for jn in range(ntor*2+1):
            n = jn-ntor
            angle = m * theta2D - nfp * n * phi2D
            sinangle = np.sin(angle)
            cosangle = np.cos(angle)
            R += RBC[m,jn] * cosangle + RBS[m,jn] * sinangle
            z += ZBC[m,jn] * cosangle + ZBS[m,jn] * sinangle

    R0 = np.zeros(nphi)
    z0 = np.zeros(nphi)
    for n in range(len(R0c)):
        angle = nfp * n * phi1D
        sinangle = np.sin(angle)
        cosangle = np.cos(angle)
        R0 += R0c[n] * cosangle + R0s[n] * sinangle
        z0 += Z0c[n] * cosangle + Z0s[n] * sinangle

    return R, z, R0, z0, r, mpol, ntor, nphi_axis

def compare_to_fortran(name, filename):
    """
    Compare output from this python code to the fortran code, for one
    of the example configurations from the papers.
    """
    # Add the directory of this file to the specified filename:
    abs_filename = os.path.join(os.path.dirname(__file__), filename)
    f      = netcdf.netcdf_file(abs_filename, 'r')
    nphi   = f.variables['N_phi'][()]
    mpol   = f.variables['mpol'][()]
    ntor   = f.variables['ntor'][()]
    r      = f.variables['r'][()]
    ntheta = 20

    py = Qsc.from_paper(name, nphi=nphi, order='r3')
    logger.info('Comparing to fortran file ' + abs_filename)

    def compare_field(fortran_name, py_field, rtol=1e-9, atol=1e-9):
        fortran_field = f.variables[fortran_name][()]
        logger.info('max difference in {}: {}'.format(fortran_name, np.max(np.abs(fortran_field - py_field))))
        np.testing.assert_allclose(fortran_field, py_field, rtol=rtol, atol=atol)

    compare_field('iota', py.iota)
    compare_field('curvature', py.curvature)
    compare_field('torsion', py.torsion)
    compare_field('sigma', py.sigma)
    compare_field('modBinv_sqrt_half_grad_B_colon_grad_B', 1 / py.L_grad_B)
    if hasattr(py, 'X20'):
        compare_field('X20', py.X20)
        compare_field('X2s', py.X2s)
        compare_field('X2c', py.X2c)
        compare_field('Y20', py.Y20)
        compare_field('Y2s', py.Y2s)
        compare_field('Y2c', py.Y2c)
        compare_field('Z20', py.Z20)
        compare_field('Z2s', py.Z2s)
        compare_field('Z2c', py.Z2c)
        compare_field('B20', py.B20)
        compare_field('d2_volume_d_psi2', py.d2_volume_d_psi2)
        compare_field('DWell_times_r2', py.DWell_times_r2)
        compare_field('DGeod_times_r2', py.DGeod_times_r2)
        compare_field('DMerc_times_r2', py.DMerc_times_r2)
        compare_field('grad_grad_B_inverse_scale_length_vs_zeta', py.grad_grad_B_inverse_scale_length_vs_varphi)
        compare_field('grad_grad_B_inverse_scale_length', py.grad_grad_B_inverse_scale_length)
        #compare_field('r_singularity', py.r_singularity) # Could be different if Newton refinement was on in 1 but not the other
        compare_field('r_singularity_basic_vs_zeta', py.r_singularity_basic_vs_varphi)
    if hasattr(py, 'X3c1'):
        compare_field('X3c1', py.X3c1)
        compare_field('X3s1', py.X3s1)
        compare_field('Y3c1', py.Y3c1)
        compare_field('Y3s1', py.Y3s1)
        compare_field('Z3c1', py.Z3c1)
        compare_field('Z3s1', py.Z3s1)
        compare_field('X3c3', py.X3c3)
        compare_field('X3s3', py.X3s3)
        compare_field('Y3c3', py.Y3c3)
        compare_field('Y3s3', py.Y3s3)
        compare_field('Z3c3', py.Z3c3)
        compare_field('Z3s3', py.Z3s3)
        compare_field('B0_order_a_squared_to_cancel', py.B0_order_a_squared_to_cancel)

    # logger.info('Creating RBC, RBS, ZBC and ZBS arrays')
    R_2D, Z_2D, _ = py.Frenet_to_cylindrical(r=r, ntheta=ntheta)
    RBC, RBS, ZBC, ZBS = to_Fourier(R_2D, Z_2D, py.nfp, mpol, ntor, py.lasym)

    RBC = RBC.transpose()
    ZBS = ZBS.transpose()
    if py.lasym:
        RBS = RBS.transpose()
        ZBC = ZBC.transpose()

    # logger.info('Comparing RBC, RBS, ZBC and ZBS arrays')
    compare_field('RBC', RBC)
    compare_field('RBS', RBS)
    compare_field('ZBC', ZBC)
    compare_field('ZBS', ZBS)

    # logger.info('Test boundary and axis splines in cylindrical coordinates')
    R_fortran, Z_fortran, R0_fortran, Z0_fortran, r, mpol, ntor, _ = fortran_plot_single(filename=filename, ntheta=ntheta, nphi=nphi)
    _, _, Z_qsc, R_qsc = py.get_boundary(r=r, ntheta=ntheta, nphi=nphi, mpol=mpol, ntor=ntor, ntheta_fourier=2*mpol)
    phi_array = np.linspace(0, 2*np.pi, nphi)
    R0_qsc = py.R0_func(phi_array)
    Z0_qsc = py.Z0_func(phi_array)
    rtol = 1e-7
    atol = 1e-7
    np.testing.assert_allclose(R_fortran, R_qsc,   rtol=rtol, atol=atol)
    np.testing.assert_allclose(Z_fortran, Z_qsc,   rtol=rtol, atol=atol)
    np.testing.assert_allclose(R0_fortran, R0_qsc, rtol=rtol, atol=atol)
    np.testing.assert_allclose(Z0_fortran, Z0_qsc, rtol=rtol, atol=atol)

    f.close()
    
class QscTests(unittest.TestCase):

    def test_curvature_torsion(self):
        """
        Test that the curvature and torsion match an independent
        calculation using the fortran code.
        """
        
        # Stellarator-symmetric case:
        stel = Qsc(rc=[1.3, 0.3, 0.01, -0.001],
                   zs=[0, 0.4, -0.02, -0.003], nfp=5, nphi=15)
        
        curvature_fortran = [1.74354628565018, 1.61776632275718, 1.5167042487094, 
                             1.9179603622369, 2.95373444883134, 3.01448808361584, 1.7714523990583, 
                             1.02055493647363, 1.02055493647363, 1.77145239905828, 3.01448808361582, 
                             2.95373444883135, 1.91796036223691, 1.5167042487094, 1.61776632275717]
        
        torsion_fortran = [0.257226801231061, -0.131225053326418, -1.12989287766591, 
                           -1.72727988032403, -1.48973327005739, -1.34398161921833, 
                           -1.76040161697108, -2.96573007082039, -2.96573007082041, 
                           -1.7604016169711, -1.34398161921833, -1.48973327005739, 
                           -1.72727988032403, -1.12989287766593, -0.13122505332643]

        varphi_fortran = [0, 0.0909479184372571, 0.181828299105257, 
                          0.268782689120682, 0.347551637441381, 0.42101745128188, 
                          0.498195826255542, 0.583626271820683, 0.673010789615233, 
                          0.758441235180374, 0.835619610154036, 0.909085423994535, 
                          0.987854372315234, 1.07480876233066, 1.16568914299866]

        rtol = 1e-13
        atol = 1e-13
        np.testing.assert_allclose(stel.curvature, curvature_fortran, rtol=rtol, atol=atol)
        np.testing.assert_allclose(stel.torsion, torsion_fortran, rtol=rtol, atol=atol)
        np.testing.assert_allclose(stel.varphi, varphi_fortran, rtol=rtol, atol=atol)

        # Non-stellarator-symmetric case:
        stel = Qsc(rc=[1.3, 0.3, 0.01, -0.001],
                   zs=[0, 0.4, -0.02, -0.003],
                   rs=[0, -0.1, -0.03, 0.002],
                   zc=[0.3, 0.2, 0.04, 0.004], nfp=5, nphi=15)
        
        curvature_fortran = [2.10743037699653, 2.33190181686696, 1.83273654023051, 
                             1.81062232906827, 2.28640008392347, 1.76919841474321, 0.919988560478029, 
                             0.741327470169023, 1.37147330126897, 2.64680884158075, 3.39786486424852, 
                             2.47005615416209, 1.50865425515356, 1.18136509189105, 1.42042418970102]
        
        torsion_fortran = [-0.167822738386845, -0.0785778346620885, -1.02205137493593, 
                           -2.05213528002946, -0.964613202459108, -0.593496282035916, 
                           -2.15852857178204, -3.72911055219339, -1.9330792779459, 
                           -1.53882290974916, -1.42156496444929, -1.11381642382793, 
                           -0.92608309386204, -0.868339812017432, -0.57696266498748]

        varphi_fortran = [0, 0.084185130335249, 0.160931495903817, 
                          0.232881563535092, 0.300551168190665, 0.368933497012765, 
                          0.444686439112853, 0.528001290336008, 0.612254611059372, 
                          0.691096975269652, 0.765820243301147, 0.846373713025902, 
                          0.941973362938683, 1.05053459351092, 1.15941650366667]
        rtol = 1e-13
        atol = 1e-13
        np.testing.assert_allclose(stel.curvature, curvature_fortran, rtol=rtol, atol=atol)
        np.testing.assert_allclose(stel.torsion, torsion_fortran, rtol=rtol, atol=atol)
        np.testing.assert_allclose(stel.varphi, varphi_fortran, rtol=rtol, atol=atol)
            
    def test_published_cases(self):
        """
        Solve the sigma equation and verify that the resulting iota matches
        published examples.
        """
        places = 7
        places2 = 3 # For max/min quantities that are less accurate
        for nphi in [50, 63]:
            # Landreman, Sengupta, Plunk (2019), section 5.1:
            stel = Qsc.from_paper('r1 section 5.1', nphi=nphi)
            self.assertEqual(stel.helicity, 0)
            self.assertAlmostEqual(stel.iota, 0.418306910215178, places=places)
            self.assertAlmostEqual(stel.max_elongation, 2.41373705531443, places=places2)
            self.assertAlmostEqual(stel.min_L_grad_B, 1 / 1.52948586064743, places=places2)
            
            # Landreman, Sengupta, Plunk (2019), section 5.2:
            stel = Qsc.from_paper('r1 section 5.2', nphi=nphi)
            self.assertEqual(stel.helicity, -1)
            self.assertAlmostEqual(stel.iota, 1.93109725535729, places=places)
            self.assertAlmostEqual(stel.max_elongation, 3.08125973323805, places=places2)
            self.assertAlmostEqual(stel.min_L_grad_B, 1 / 4.73234243198959, places=places2)
            
            # Landreman, Sengupta, Plunk (2019), section 5.3:
            stel = Qsc.from_paper('r1 section 5.3', nphi=nphi)
            self.assertEqual(stel.helicity, 0)
            self.assertAlmostEqual(stel.iota, 0.311181373123728, places=places)
            self.assertAlmostEqual(stel.max_elongation, 3.30480616121377, places=places2)
            self.assertAlmostEqual(stel.min_L_grad_B, 1 / 1.7014044379421, places=places2)
        
            # Landreman & Sengupta (2019), section 5.1:
            stel = Qsc.from_paper('r2 section 5.1', nphi=nphi)
            self.assertEqual(stel.helicity, 0)
            self.assertAlmostEqual(stel.iota, -0.420473351810416 , places=places)
            self.assertAlmostEqual(stel.max_elongation, 4.38384260252044, places=places2)
            self.assertAlmostEqual(stel.min_L_grad_B, 1 / 1.39153088147691, places=places2)
            check_r2(stel)
        
            # Landreman & Sengupta (2019), section 5.2:
            stel = Qsc.from_paper('r2 section 5.2', nphi=nphi)
            self.assertEqual(stel.helicity, 0)
            self.assertAlmostEqual(stel.iota, -0.423723995700502, places=places)
            self.assertAlmostEqual(stel.max_elongation, 4.86202324600918, places=places2)
            self.assertAlmostEqual(stel.min_L_grad_B, 1 / 1.47675199709439, places=places2)
            check_r2(stel)
            
            # Landreman & Sengupta (2019), section 5.3:
            stel = Qsc.from_paper('r2 section 5.3', nphi=nphi)
            self.assertEqual(stel.helicity, 0)
            self.assertAlmostEqual(stel.iota, 0.959698159859113, places=places)
            self.assertAlmostEqual(stel.max_elongation, 2.20914173760329, places=places2)
            self.assertAlmostEqual(stel.min_L_grad_B, 1 / 1.4922510395338, places=places2)
            check_r2(stel)
            
            # Landreman & Sengupta (2019), section 5.4:
            stel = Qsc.from_paper('r2 section 5.4', nphi=nphi)
            self.assertEqual(stel.helicity, 1)
            self.assertAlmostEqual(stel.iota, -1.14413695118515, places=places)
            self.assertAlmostEqual(stel.max_elongation, 2.98649978627541, places=places2)
            self.assertAlmostEqual(stel.min_L_grad_B, 1 / 2.64098280647292, places=places2)
            check_r2(stel)
            
            # Landreman & Sengupta (2019), section 5.5:
            stel = Qsc.from_paper('r2 section 5.5', nphi=nphi)
            self.assertEqual(stel.helicity, 1)
            self.assertAlmostEqual(stel.iota, -0.828885267089981, places=places)
            self.assertAlmostEqual(stel.max_elongation, 3.6226360623368, places=places2)
            self.assertAlmostEqual(stel.min_L_grad_B, 1 / 4.85287603883526, places=places2)
            check_r2(stel)

    def test_QI_r2(self):
        rc      = [ 1.0,0.0,-0.2,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0 ]
        zs      = [ 0.0,0.0,0.26443390548300216,0.0,-0.01566057704554211,0.0,0.0022396012358901997,0.0,-0.0009650267041755669,0.0,7.207283882530656e-05,0.0,3.775829720639504e-05,0.0,-1.3146652962906995e-05,0.0,3.1591666411150446e-06 ]
        B0_vals = [ 1.0,0.21379635338650388 ]
        omn_method ='buffer'
        k_buffer = 2
        k_second_order_SS   = 6.557610594867123
        d_svals = [0.]
        delta   = 1.3316877731759529
        nfp     = 1
        B2s_svals = [ 0.0,0.12356397260960127,-0.3703624809486158,-0.03495724243250028,-0.024264380562857745,-0.2363887868868969,-0.05181276831340249,0.08138881214364944,0.005241997297581 ]
        B2c_cvals = [ -0.40251233990996754,0.10398784730800278,0.5170167340215038,-0.2255734311522203,0.044645330385214954,0.21871024881966092,-0.06764902501780555,0.014096798551213216,-0.05415701995955931,0.0 ]
        p2      =  0.0
        nphi    =  301
        stel    =  Qsc(omn_method = omn_method, k_buffer=k_buffer, rc=rc,zs=zs, nfp=nfp, B0_vals=B0_vals, d_svals=d_svals, nphi=nphi, omn=True, delta=delta, B2c_cvals=B2c_cvals, B2s_svals=B2s_svals, p2=p2, order='r3', k_second_order_SS=k_second_order_SS)
        check_r2(stel)

    def test_compare_to_fortran(self):
        """
        Compare the output of this python code to the fortran code.
        """
        compare_to_fortran("r2 section 5.1", "quasisymmetry_out.LandremanSengupta2019_section5.1.nc")
        compare_to_fortran("r2 section 5.2", "quasisymmetry_out.LandremanSengupta2019_section5.2.nc")
        compare_to_fortran("r2 section 5.3", "quasisymmetry_out.LandremanSengupta2019_section5.3.nc")
        compare_to_fortran("r2 section 5.4", "quasisymmetry_out.LandremanSengupta2019_section5.4.nc")
        compare_to_fortran("r2 section 5.5", "quasisymmetry_out.LandremanSengupta2019_section5.5.nc")

    def test_change_nfourier(self):
        """
        Test the change_nfourier() method.
        """
        rtol = 1e-13
        atol = 1e-13
        s1 = Qsc.from_paper('r2 section 5.2')
        m = s1.nfourier
        for n in range(2, 7):
            s2 = Qsc.from_paper('r2 section 5.2')
            s2.change_nfourier(n)
            if n <= m:
                # We lowered nfourier
                np.testing.assert_allclose(s1.rc[:n], s2.rc, rtol=rtol, atol=atol)
                np.testing.assert_allclose(s1.rs[:n], s2.rs, rtol=rtol, atol=atol)
                np.testing.assert_allclose(s1.zc[:n], s2.zc, rtol=rtol, atol=atol)
                np.testing.assert_allclose(s1.zs[:n], s2.zs, rtol=rtol, atol=atol)
            else:
                # We increased nfourier
                np.testing.assert_allclose(s1.rc, s2.rc[:m], rtol=rtol, atol=atol)
                np.testing.assert_allclose(s1.rs, s2.rs[:m], rtol=rtol, atol=atol)
                np.testing.assert_allclose(s1.zc, s2.zc[:m], rtol=rtol, atol=atol)
                np.testing.assert_allclose(s1.zs, s2.zs[:m], rtol=rtol, atol=atol)
                z = np.zeros(n - s1.nfourier)
                np.testing.assert_allclose(z, s2.rc[m:], rtol=rtol, atol=atol)
                np.testing.assert_allclose(z, s2.rs[m:], rtol=rtol, atol=atol)
                np.testing.assert_allclose(z, s2.zc[m:], rtol=rtol, atol=atol)
                np.testing.assert_allclose(z, s2.zs[m:], rtol=rtol, atol=atol)
                
if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import logging
import os
import unittest
import numpy as np
from scipy.io import netcdf
import matplotlib.pyplot as plt
from qsc.qsc import Qsc

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

cases = ['r1 section 5.1', 'r1 section 5.2', 'r1 section 5.3',
         'r2 section 5.1', 'r2 section 5.2', 'r2 section 5.3', 'r2 section 5.4', 'r2 section 5.5']

# Calling plt.close() after each matplotlib figure is not essential,
# but it eliminates a matplotlib warning about opening too many
# figures at once, and it reduces memory usage.

class PlotTests(unittest.TestCase):

    def test_plot(self):
        for case in cases:
            s = Qsc.from_paper(case, order='r1')
            s.plot(show=False)
            plt.close()
        
            s = Qsc.from_paper(case, order='r2')
            s.plot(show=False)
            plt.close()
        
            s = Qsc.from_paper(case, order='r3')
            s.plot(show=False)
            plt.close()
        
    def test_axis(self):
        """
        Test call to plot axis shape
        """
        stel = Qsc.from_paper('r1 section 5.2')
        stel.plot_axis(frenet=False, show=False)
        plt.close()

        stel = Qsc.from_paper(4, order='r2')
        stel.plot_axis(frenet=False, show=False)
        plt.close()

        stel = Qsc.from_paper(4, order='r3')
        stel.plot_axis(frenet=False, show=False)
        plt.close()
                
    def test_other_plots(self):
        """
        Call other plotting functions to check that they work with a first
        order case and a second order one
        """
        stel = Qsc.from_paper("r1 section 5.1")
        # stel.plot(fieldlines=True)
        stel.plot_boundary(show=False)
        stel.B_fieldline(show=False)
        plt.close()
        stel.B_contour(show=False)
        plt.close()
        
        stel = Qsc.from_paper(4, order='r2')
        # stel.plot(fieldlines=True)
        stel.plot_boundary(show=False)
        stel.B_fieldline(show=False)
        plt.close()
        stel.B_contour(show=False)
        plt.close()
        
        stel = Qsc.from_paper(4, order='r3')
        # stel.plot(fieldlines=True)
        stel.plot_boundary(show=False)
        stel.B_fieldline(show=False)
        plt.close()
        stel.B_contour(show=False)
        plt.close()

if __name__ == "__main__":
    unittest.main()
